/*
 * contactview.cpp - contact list widget
 * Copyright (C) 2001, 2002  Justin Karneges
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include"contactview.h"

#include<qapplication.h>
#include<qaction.h>
#include<qptrlist.h>
#include<qheader.h>
#include<qtimer.h>
#include<qpainter.h>
#include<qpopupmenu.h>
#include<qmessagebox.h>
#include<qinputdialog.h>
#include<qiconset.h>
#include<qdragobject.h>
#include"kmp/types.h"
#include"common.h"
#include"userlist.h"
#include"psiaccount.h"
#include"iconset.h"

//----------------------------------------------------------------------------
// ContactProfile
//----------------------------------------------------------------------------
class ContactProfile::Entry
{
public:
	Entry()
	{
		alerting = false;
		cvi.setAutoDelete(true);
	}
	~Entry()
	{
	}

	UserListItem u;
	QPtrList<ContactViewItem> cvi;
	bool alerting;
	Anim *anim;
};

class ContactProfile::Private
{
public:
	Private() {}

	QString name;
	ContactView *cv;
	ContactViewItem *cvi;
	ContactViewItem *self;
	UserListItem su;
	QPtrList<Entry> roster;
	QPtrList<ContactViewItem> groups;
	int oldstate;
	QTimer *t;
	PsiAccount *pa;
};

ContactProfile::ContactProfile(PsiAccount *pa, const QString &name, ContactView *cv, bool unique)
{
	d = new Private;
	d->pa = pa;
	d->name = name;
	d->cv = cv;
	d->cv->link(this);
	d->t = new QTimer;
	connect(d->t, SIGNAL(timeout()), SLOT(updateGroups()));

	d->roster.setAutoDelete(true);

	d->self = 0;

	if(!unique)
		d->cvi = new ContactViewItem(name, this, d->cv);
	else
		d->cvi = 0;

	d->oldstate = -2;

	deferredUpdateGroups();
}

ContactProfile::~ContactProfile()
{
	// delete the roster
	clear();

	// clean up
	delete d->self;
	delete d->cvi;

	delete d->t;
	d->cv->unlink(this);

	delete d;
}

ContactView *ContactProfile::contactView() const
{
	return d->cv;
}

ContactViewItem *ContactProfile::self() const
{
	return d->self;
}

const QString & ContactProfile::name() const
{
	return d->name;
}

void ContactProfile::setName(const QString &name)
{
	d->name = name;
	if(d->cvi)
		d->cvi->setProfileName(name);
}

void ContactProfile::setState(int state)
{
	if(state == d->oldstate)
		return;
	d->oldstate = state;

	if(d->cvi) {
		d->cv->resetAnim();
		d->cvi->setProfileState(state);
	}
}

void ContactProfile::setUsingSSL(bool on)
{
	if(d->cvi)
		d->cvi->setProfileSSL(on);
}

ContactViewItem *ContactProfile::addGroup(int type)
{
	ContactViewItem *item;

	QString gname;
	if(type == ContactViewItem::gGeneral)
		gname = tr("General");
	else if(type == ContactViewItem::gNotInList)
		gname = tr("Not in list");
	else if(type == ContactViewItem::gAgents)
		gname = tr("Agents/Transports");

	if(d->cvi)
		item = new ContactViewItem(gname, type, this, d->cvi);
	else
		item = new ContactViewItem(gname, type, this, d->cv);

	if(type == ContactViewItem::gAgents && !d->cv->isShowAgents())
		item->setVisible(false);

	d->groups.append(item);

	return item;
}

ContactViewItem *ContactProfile::addGroup(const QString &name)
{
	ContactViewItem *item;
	if(d->cvi)
		item = new ContactViewItem(name, ContactViewItem::gUser, this, d->cvi);
	else
		item = new ContactViewItem(name, ContactViewItem::gUser, this, d->cv);

	d->groups.append(item);

	return item;
}

// check for special group
ContactViewItem *ContactProfile::checkGroup(int type)
{
	ContactViewItem *item;
	if(d->cvi)
		item = (ContactViewItem *)d->cvi->firstChild();
	else
		item = (ContactViewItem *)d->cv->firstChild();

	for(; item; item = (ContactViewItem *)item->nextSibling()) {
		if(item->type() == ContactViewItem::Group && item->groupType() == type)
			return item;
	}

	return 0;
}

// check for user group
ContactViewItem *ContactProfile::checkGroup(const QString &name)
{
	ContactViewItem *item;
	if(d->cvi)
		item = (ContactViewItem *)d->cvi->firstChild();
	else
		item = (ContactViewItem *)d->cv->firstChild();

	for(; item; item = (ContactViewItem *)item->nextSibling()) {
		if(item->type() == ContactViewItem::Group && item->groupType() == ContactViewItem::gUser && item->groupName() == name)
				return item;
	}

	return 0;
}

ContactViewItem *ContactProfile::ensureGroup(int type)
{
	ContactViewItem *group_item = checkGroup(type);
	if(!group_item)
		group_item = addGroup(type);

	return group_item;
}

ContactViewItem *ContactProfile::ensureGroup(const QString &name)
{
	ContactViewItem *group_item = checkGroup(name);
	if(!group_item)
		group_item = addGroup(name);

	return group_item;
}

void ContactProfile::checkDestroyGroup(const QString &group)
{
	ContactViewItem *group_item = checkGroup(group);
	if(group_item)
		checkDestroyGroup(group_item);
}

void ContactProfile::checkDestroyGroup(ContactViewItem *group)
{
	if(group->childCount() == 0) {
		d->groups.remove(group);
		delete group;
	}
}

void ContactProfile::updateEntry(const UserListItem &u)
{
	Entry *e = findEntry(u.jid());
	if(!e) {
		e = new Entry;
		d->roster.append(e);
		e->u = u;
	}
	else {
		e->u = u;
		removeUnneededContactItems(e);

		// update remaining items
		QPtrListIterator<ContactViewItem> it(e->cvi);
		for(ContactViewItem *i; (i = it.current()); ++it) {
			i->setContact(&e->u);
			if(!u.isAvailable())
				i->stopAnimateNick();
		}
	}

	deferredUpdateGroups();
	addNeededContactItems(e);
}

void ContactProfile::updateSelf(const UserListItem &u)
{
	d->su = u;
	if(!d->self) {
		if(!d->cvi)
			return;
		d->self = new ContactViewItem(&d->su, this, d->cvi);
	}
	else {
		d->self->setContact(&d->su);
		if(!u.isAvailable())
			d->self->stopAnimateNick();
	}
}

void ContactProfile::removeSelf()
{
	delete d->self;
	d->self = 0;
}

ContactViewItem *ContactProfile::addContactItem(Entry *e, ContactViewItem *group_item)
{
	ContactViewItem *i = new ContactViewItem(&e->u, this, group_item);
	e->cvi.append(i);
	if(e->alerting)
		i->setAlert(e->anim);
	deferredUpdateGroups();
	//printf("ContactProfile: adding [%s] to group [%s]\n", e->u.jid().full().latin1(), group_item->groupName().latin1());
	return i;
}

ContactViewItem *ContactProfile::ensureContactItem(Entry *e, ContactViewItem *group_item)
{
	QPtrListIterator<ContactViewItem> it(e->cvi);
	for(ContactViewItem *i; (i = it.current()); ++it) {
		ContactViewItem *g = (ContactViewItem *)i->parent();
		if(g == group_item)
			return i;
	}
	return addContactItem(e, group_item);
}

void ContactProfile::removeContactItem(Entry *e, ContactViewItem *i)
{
	ContactViewItem *group_item = (ContactViewItem *)i->parent();
	//printf("ContactProfile: removing [%s] from group [%s]\n", e->u.jid().full().latin1(), group_item->groupName().latin1());
	e->cvi.removeRef(i);
	deferredUpdateGroups();
	checkDestroyGroup(group_item);
}

void ContactProfile::addNeededContactItems(Entry *e)
{
	const UserListItem &u = e->u;

	if(u.inList()) {
		// don't add if we're not supposed to see it
		if(u.isTransport()) {
			if(!d->cv->isShowAgents())
				return;
		}
		else {
			if(!e->alerting) {
				if((!d->cv->isShowOffline() && !u.isAvailable()) || (!d->cv->isShowAway() && u.isAway()))
					return;
			}
		}
	}

	if(!u.inList())
		ensureContactItem(e, ensureGroup(ContactViewItem::gNotInList));
	else if(u.isTransport())
		ensureContactItem(e, ensureGroup(ContactViewItem::gAgents));
	else if(u.groups().isEmpty())
		ensureContactItem(e, ensureGroup(ContactViewItem::gGeneral));
	else {
		const QStringList &groups = u.groups();
		for(QStringList::ConstIterator git = groups.begin(); git != groups.end(); ++git)
			ensureContactItem(e, ensureGroup(*git));
	}
}

void ContactProfile::removeUnneededContactItems(Entry *e)
{
	const UserListItem &u = e->u;

	if(u.inList()) {
		bool delAll = false;
		if(u.isTransport()) {
			if(!d->cv->isShowAgents())
				delAll = true;
		}
		else {
			if(!e->alerting) {
				if((!d->cv->isShowOffline() && !u.isAvailable()) || (!d->cv->isShowAway() && u.isAway()))
					delAll = true;
			}
		}
		if(delAll) {
			clearContactItems(e);
			return;
		}
	}

	QPtrListIterator<ContactViewItem> it(e->cvi);
	for(ContactViewItem *i; (i = it.current());) {
		bool del = false;
		ContactViewItem *g = (ContactViewItem *)i->parent();

		if(g->groupType() == ContactViewItem::gNotInList && u.inList())
			del = true;
		else if(g->groupType() != ContactViewItem::gNotInList && !u.inList())
			del = true;
		else if(g->groupType() == ContactViewItem::gAgents && !u.isTransport())
			del = true;
		else if(g->groupType() != ContactViewItem::gAgents && u.isTransport())
			del = true;
		else if(g->groupType() == ContactViewItem::gGeneral && !u.groups().isEmpty())
			del = true;
		else if(g->groupType() != ContactViewItem::gGeneral && u.groups().isEmpty() && !u.isTransport())
			del = true;
		else if(g->groupType() == ContactViewItem::gUser) {
			const QStringList &groups = u.groups();
			if(!groups.isEmpty()) {
				bool found = false;
				for(QStringList::ConstIterator git = groups.begin(); git != groups.end(); ++git) {
					if(g->groupName() == *git) {
						found = true;
						break;
					}
				}
				if(!found)
					del = true;
			}
		}

		if(del)
			removeContactItem(e, i);
		else
			++it;
	}
}

void ContactProfile::clearContactItems(Entry *e)
{
	QPtrListIterator<ContactViewItem> it(e->cvi);
	for(ContactViewItem *i; (i = it.current());)
		removeContactItem(e, i);
}

void ContactProfile::addAllNeededContactItems()
{
	QPtrListIterator<Entry> it(d->roster);
	for(Entry *e; (e = it.current()); ++it)
		addNeededContactItems(e);
}

void ContactProfile::removeAllUnneededContactItems()
{
	QPtrListIterator<Entry> it(d->roster);
	for(Entry *e; (e = it.current()); ++it)
		removeUnneededContactItems(e);
}

void ContactProfile::removeEntry(const Jid &j)
{
	Entry *e = findEntry(j);
	if(e)
		removeEntry(e);
}

void ContactProfile::removeEntry(Entry *e)
{
	e->alerting = false;
	clearContactItems(e);
	d->roster.remove(e);
}

void ContactProfile::setAlert(const Jid &j, Anim *type)
{
	// reset all animations (so they are in sync with this new one)
	d->cv->resetAnim();

	if(d->su.jid().compare(j)) {
		d->self->setAlert(type);
	}
	else {
		Entry *e = findEntry(j);
		if(!e)
			return;

		e->alerting = true;
		e->anim = type;
		addNeededContactItems(e);
		QPtrListIterator<ContactViewItem> it(e->cvi);
		for(ContactViewItem *i; (i = it.current()); ++it)
			i->setAlert(type);

		if(option.scrollTo)
			ensureVisible(e);
	}
}

void ContactProfile::clearAlert(const Jid &j)
{
	if(d->su.jid().compare(j)) {
		d->self->clearAlert();
	}
	else {
		Entry *e = findEntry(j);
		if(!e)
			return;

		e->alerting = false;
		QPtrListIterator<ContactViewItem> it(e->cvi);
		for(ContactViewItem *i; (i = it.current()); ++it)
			i->clearAlert();
		if(e->u.inList())
			removeUnneededContactItems(e);
	}
}

void ContactProfile::clear()
{
	QPtrListIterator<Entry> it(d->roster);
	for(Entry *e; (e = it.current());)
		removeEntry(e);
}

ContactProfile::Entry *ContactProfile::findEntry(const Jid &jid) const
{
	QPtrListIterator<Entry> it(d->roster);
	for(Entry *e; (e = it.current()); ++it) {
		if(e->u.jid().compare(jid))
			return e;
	}
	return 0;
}

ContactProfile::Entry *ContactProfile::findEntry(ContactViewItem *i) const
{
	QPtrListIterator<Entry> it(d->roster);
	for(Entry *e; (e = it.current()); ++it) {
		QPtrListIterator<ContactViewItem> ci(e->cvi);
		for(ContactViewItem *cvi; (cvi = ci.current()); ++ci) {
			if(cvi == i)
				return e;
		}
	}
	return 0;
}

// return a list of contacts from a CVI group
JidList ContactProfile::contactListFromCVGroup(ContactViewItem *group) const
{
	JidList list;

	for(ContactViewItem *item = (ContactViewItem *)group->firstChild(); item ; item = (ContactViewItem *)item->nextSibling()) {
		if(item->type() != ContactViewItem::Contact)
			continue;

		list.append(item->u()->jid());
	}

	return list;
}

// return the number of contacts from a CVI group
int ContactProfile::contactSizeFromCVGroup(ContactViewItem *group) const
{
	int total = 0;

	for(ContactViewItem *item = (ContactViewItem *)group->firstChild(); item ; item = (ContactViewItem *)item->nextSibling()) {
		if(item->type() != ContactViewItem::Contact)
			continue;

		++total;
	}

	return total;
}

// return the number of contacts from a CVI group
int ContactProfile::contactsOnlineFromCVGroup(ContactViewItem *group) const
{
	int total = 0;

	for(ContactViewItem *item = (ContactViewItem *)group->firstChild(); item ; item = (ContactViewItem *)item->nextSibling()) {
		if(item->type() == ContactViewItem::Contact && item->u()->isAvailable())
			++total;
	}

	return total;
}

// return a list of contacts associated with "groupName"
JidList ContactProfile::contactListFromGroup(const QString &groupName) const
{
	JidList list;

	QPtrListIterator<Entry> it(d->roster);
	for(Entry *e; (e = it.current()); ++it) {
		const UserListItem &u = e->u;
		if(u.isTransport())
			continue;
		const QStringList &g = u.groups();
		if(g.isEmpty()) {
			if(groupName.isEmpty())
				list.append(u.jid());
		}
		else {
			for(QStringList::ConstIterator git = g.begin(); git != g.end(); ++git) {
				if(*git == groupName) {
					list.append(u.jid());
					break;
				}
			}
		}
	}

	return list;
}

// return the number of contacts associated with "groupName"
int ContactProfile::contactSizeFromGroup(const QString &groupName) const
{
	int total = 0;

	QPtrListIterator<Entry> it(d->roster);
	for(Entry *e; (e = it.current()); ++it) {
		const UserListItem &u = e->u;
		if(u.isTransport())
			continue;
		const QStringList &g = u.groups();
		if(g.isEmpty()) {
			if(groupName.isEmpty())
				++total;
		}
		else {
			for(QStringList::ConstIterator git = g.begin(); git != g.end(); ++git) {
				if(*git == groupName) {
					++total;
					break;
				}
			}
		}
	}

	return total;
}

void ContactProfile::updateGroupInfo(ContactViewItem *group)
{
	int type = group->groupType();
	if(type == ContactViewItem::gGeneral || type == ContactViewItem::gAgents || type == ContactViewItem::gUser) {
		int online = contactsOnlineFromCVGroup(group);
		int total;
		QString gname;
		if(type == ContactViewItem::gUser)
			gname = group->groupName();
		else
			gname = "";
		total = contactSizeFromGroup(gname);

		group->setGroupInfo(QString("(%1/%2)").arg(online).arg(total));
	}
	else {
		int inGroup = contactSizeFromCVGroup(group);
		group->setGroupInfo(QString("(%1)").arg(inGroup));
	}
}

QStringList ContactProfile::groupList() const
{
	QStringList groupList;

	QPtrListIterator<Entry> it(d->roster);
	for(Entry *e; (e = it.current()); ++it) {
		const QStringList &groups = e->u.groups();
		for(QStringList::ConstIterator git = groups.begin(); git != groups.end(); ++git) {
			if(qstringlistmatch(groupList, *git) == -1)
				groupList.append(*git);
		}
	}

	groupList.sort();
	return groupList;
}

void ContactProfile::animateNick(const Jid &j)
{
	if(d->su.jid().compare(j))
		d->self->setAnimateNick();

	Entry *e = findEntry(j);
	if(!e)
		return;
	QPtrListIterator<ContactViewItem> it(e->cvi);
	for(ContactViewItem *i; (i = it.current()); ++it)
		i->setAnimateNick();
}

void ContactProfile::deferredUpdateGroups()
{
	d->t->start(250, true);
}

void ContactProfile::updateGroups()
{
	int totalOnline = 0;
	{
		QPtrListIterator<Entry> it(d->roster);
		for(Entry *e; (e = it.current()); ++it) {
			if(e->u.isAvailable())
				++totalOnline;
		}
		if(d->cvi)
			d->cvi->setGroupInfo(QString("(%1/%2)").arg(totalOnline).arg(d->roster.count()));
	}

	{
		QPtrListIterator<ContactViewItem> it(d->groups);
		for(ContactViewItem *g; (g = it.current()); ++it)
			updateGroupInfo(g);
	}
}

void ContactProfile::ensureVisible(const Jid &j)
{
	Entry *e = findEntry(j);
	if(!e)
		return;
	ensureVisible(e);
}

void ContactProfile::ensureVisible(Entry *e)
{
	if(!e->alerting) {
		if(!d->cv->isShowAgents() && e->u.isTransport())
			d->cv->setShowAgents(true);
		if(!d->cv->isShowOffline() && !e->u.isAvailable())
			d->cv->setShowOffline(true);
		if(!d->cv->isShowAway() && e->u.isAway())
			d->cv->setShowAway(true);
	}

	ContactViewItem *i = e->cvi.first();
	if(!i)
		return;
	d->cv->ensureItemVisible(i);
}

void ContactProfile::doContextMenu(ContactViewItem *i, const QPoint &pos)
{
	bool online = d->pa->loggedIn();

	if(i->type() == ContactViewItem::Profile) {
		QPopupMenu pm;

		QPopupMenu *am = new QPopupMenu(&pm);
		am->insertItem(QIconSet(is.send), tr("Send server message"), 1);
		am->insertItem(QIconSet(is.send), tr("Set MOTD"), 2);

		pm.insertItem(QIconSet(is.status(STATUS_ONLINE)),  status2txt(STATUS_ONLINE),      STATUS_ONLINE +4);
		pm.insertItem(QIconSet(is.status(STATUS_AWAY)),    status2txt(STATUS_AWAY),        STATUS_AWAY   +4);
		pm.insertItem(QIconSet(is.status(STATUS_XA)),      status2txt(STATUS_XA),          STATUS_XA     +4);
		pm.insertItem(QIconSet(is.status(STATUS_DND)),     status2txt(STATUS_DND),         STATUS_DND    +4);
		pm.insertSeparator();
		pm.insertItem(QIconSet(is.status(STATUS_OFFLINE)), status2txt(STATUS_OFFLINE),     STATUS_OFFLINE+4);


		pm.insertSeparator();
		pm.insertItem(tr("Modify Account..."), 0);
		pm.insertSeparator();
		pm.insertItem(tr("&Admin"), am);

		int x = pm.exec(pos);

		if(x == -1)
			return;

		if(x == 0)
			d->pa->modify();
		else if(x == 1) {
			Jid j = d->pa->jid().host() + '/' + "announce/online";
			actionSendMessage(j);
		}
		else if(x == 2) {
			Jid j = d->pa->jid().host() + '/' + "announce/motd";
			actionSendMessage(j);
		}
		else if(x >= 4) {
			int status = x - 4;
			d->pa->changeStatus(status);
		}
	}
	else if(i->type() == ContactViewItem::Group) {
		QString gname = i->groupName();
		QPopupMenu pm;

		pm.insertItem(QIconSet(is.send), tr("Send message to group"), 0);
		d->cv->qa_ren->addTo(&pm);
		pm.insertSeparator();
		pm.insertItem(QIconSet(is.remove), tr("Remove group"), 2);
		pm.insertItem(QIconSet(is.remove), tr("Remove group and contacts"), 3);

		// disable if it's not a user group
		if(!online || i->groupType() != ContactViewItem::gUser) {
			d->cv->qa_ren->setEnabled(false);
			pm.setItemEnabled(2, false);
			pm.setItemEnabled(3, false);
		}
		else
			d->cv->qa_ren->setEnabled(true);

		if(i->groupType() == ContactViewItem::gAgents) {
			pm.insertSeparator();
			pm.insertItem(tr("Hide"), 4);
		}

		int x = pm.exec(pos);

		if(x == -1)
			return;

		if(x == 0) {
			JidList list = contactListFromCVGroup(i);

			// send multi
			actionSendMessage(list);
		}
		else if(x == 2 && online) {
			int n = QMessageBox::information(d->cv, tr("Remove Group"),tr(
			"This will cause all contacts in this group to be disassociated with it.\n"
			"\n"
			"Proceed?"), tr("&No"), tr("&Yes"));

			if(n == 1) {
				JidList list = contactListFromGroup(i->groupName());
				for(QValueList<Jid>::Iterator it = list.begin(); it != list.end(); ++it)
					actionGroupRemove(*it, gname);
			}
		}
		else if(x == 3 && online) {
			int n = QMessageBox::information(d->cv, tr("Remove Group and Contacts"),tr(
			"WARNING!  This will remove all contacts associated with this group!\n"
			"\n"
			"Proceed?"), tr("&No"), tr("&Yes"));

			if(n == 1) {
				JidList list = contactListFromGroup(i->groupName());
				for(QValueList<Jid>::Iterator it = list.begin(); it != list.end(); ++it) {
					removeEntry(*it);
					actionRemove(*it);
				}
			}
		}
		else if(x == 4) {
			if(i->groupType() == ContactViewItem::gAgents)
				d->cv->setShowAgents(false);
		}
	}
	else if(i->type() == ContactViewItem::Contact) {
		bool self = false;
		UserListItem *u;
		Entry *e = 0;
		if(i == d->self) {
			self = true;
			u = &d->su;
		}
		else {
			e = findEntry(i);
			if(!e)
				return;
			u = &e->u;
		}
		QStringList gl = groupList();
		bool inList = e ? e->u.inList() : false;
		bool isAgent = e ? e->u.isTransport() : false;
		bool avail = e ? e->u.isAvailable() : false;

		QPopupMenu pm;

		if(!self && !inList) {
			pm.insertItem(QIconSet(is.add), tr("Add/Authorize to contact list"), 10);
			if(!online)
				pm.setItemEnabled(10, false);
			pm.insertSeparator();
		}

		if(self) {
			if(i->isAlerting()) {
				d->cv->qa_recv->addTo(&pm);
				pm.insertSeparator();
			}
		}
		else {
			if(e->alerting) {
				d->cv->qa_recv->addTo(&pm);
				pm.insertSeparator();
			}
		}

		d->cv->qa_send->setIconSet(is.send.pixmap());
		d->cv->qa_send->addTo(&pm);

		//pm.insertItem(QIconSet(is.url), tr("Send &URL"), 2);

		const UserResourceList &rl = u->userResourceList();

		int base_sendto = 32;
		int at_sendto = 0;
		QPopupMenu *s2m = new QPopupMenu(&pm);
		QPopupMenu *c2m = new QPopupMenu(&pm);
		//s2m->insertItem(tr("Recipient Default"), base_sendto+at_sendto++);
		//c2m->insertItem(tr("Recipient Default"), base_sendto+at_sendto++);
		if(!rl.isEmpty()) {
			//s2m->insertSeparator();
			//c2m->insertSeparator();
			for(UserResourceList::ConstIterator it = rl.begin(); it != rl.end(); ++it) {
				const UserResource &r = *it;
				QString rname;
				if(r.name().isEmpty())
					rname = tr("[blank]");
				else
					rname = r.name();
				s2m->insertItem(QIconSet(is.status(r.status())), rname, base_sendto+at_sendto++);
				c2m->insertItem(QIconSet(is.status(r.status())), rname, base_sendto+at_sendto++);
			}
		}
		pm.insertItem(tr("Send message to"), s2m, 17);

		d->cv->qa_chat->setIconSet(is.chat.base().pixmap());
		d->cv->qa_chat->addTo(&pm);

		pm.insertItem(tr("Open chat to"), c2m, 18);

		if(rl.isEmpty()) {
			pm.setItemEnabled(17, false);
			pm.setItemEnabled(18, false);
		}

		int base_hidden = base_sendto + at_sendto;
		int at_hidden = 0;
		QStringList hc = d->pa->hiddenChats(u->jid());
		QPopupMenu *cm = new QPopupMenu(&pm);
		for(QStringList::ConstIterator it = hc.begin(); it != hc.end(); ++it) {
			QString rname;
			if((*it).isEmpty())
				rname = tr("[blank]");
			else
				rname = *it;

			// determine status
			int status;
			const UserResourceList &rl = u->userResourceList();
			UserResourceList::ConstIterator uit = rl.find(*it);
			if(uit == rl.end())
				status = STATUS_OFFLINE;
			else
				status = makeSTATUS((*uit).status());
			cm->insertItem(QIconSet(is.status(status)), rname, base_hidden+at_hidden++);
		}
		pm.insertItem(tr("Active chats"), cm, 7);
		if(hc.isEmpty())
			pm.setItemEnabled(7, false);

		pm.insertSeparator();

		int base_group = base_hidden + at_hidden;

		if(!self) {
			if(inList) {
				d->cv->qa_ren->addTo(&pm);
				d->cv->qa_ren->setEnabled(online);
			}

			if(!isAgent) {
				if(inList) {
					QPopupMenu *gm = new QPopupMenu(&pm);

					gm->setCheckable(true);
					gm->insertItem(tr("&None"), 8);
					gm->insertSeparator();

					QString g;
					if(e->u.groups().isEmpty())
						gm->setItemChecked(8, true);
					else
						g = ((ContactViewItem *)i->parent())->groupName();

					int n = 0;
					for(QStringList::ConstIterator it = gl.begin(); it != gl.end(); ++it) {
						QString str;
						if(n < 9)
							str = "&";
						str += QString("%1. %2").arg(n+1).arg(*it);
						gm->insertItem(str, n+base_group);

						if(*it == g)
							gm->setItemChecked(n+base_group, true);
						++n;
					}
					if(n > 0)
						gm->insertSeparator();

					gm->insertItem(tr("&Create new..."), 9);
					pm.insertItem(tr("&Group"), gm, 5);

					if(!online)
						pm.setItemEnabled(5, false);
				}
			}
			else {
				pm.insertSeparator();

				if(!avail && online)
					d->cv->qa_logon->setEnabled(true);
				else
					d->cv->qa_logon->setEnabled(false);

				d->cv->qa_logon->setIconSet(transport2icon(e->u.jid().full(), STATUS_ONLINE));
				d->cv->qa_logon->addTo(&pm);

				pm.insertItem(transport2icon(e->u.jid().full(), STATUS_OFFLINE), tr("Log off"), 16);
				if(!avail || !online)
					pm.setItemEnabled(16, false);
				pm.insertSeparator();
			}
		}

		if(inList) {
			pm.insertItem(tr("Resend authorization to"), 6);
			pm.insertItem(tr("Rerequest authorization from"), 11);
			pm.insertItem(tr("Remove authorization from"), 15);
			if(!online) {
				pm.setItemEnabled(6, false);
				pm.setItemEnabled(11, false);
				pm.setItemEnabled(15, false);
			}
		}

		if(!self) {
			if(online || !inList)
				d->cv->qa_rem->setEnabled(true);
			else
				d->cv->qa_rem->setEnabled(false);

			d->cv->qa_rem->setIconSet(is.remove.pixmap());
			d->cv->qa_rem->addTo(&pm);
			pm.insertSeparator();
		}

		pm.insertItem(tr("Check &Status"), 4);

		pm.insertItem(QIconSet(is.info), tr("User &Info"), 12);

		d->cv->qa_hist->setIconSet(is.history.pixmap());
		d->cv->qa_hist->addTo(&pm);

		if(dtcp_use) {
			pm.insertSeparator();
			pm.insertItem(tr("DTCP Test"), 13);
		}

		QString name = u->jid().full();
		QString show = jidnick(u->jid().full(), u->name());
		if(name != show)
			name += QString(" (%1)").arg(u->name());

		int x = pm.exec(pos);

		if(x == -1)
			return;

		if(x == 0) {
			actionRecvEvent(u->jid());
		}
		else if(x == 1) {
			actionSendMessage(u->jid());
		}
		else if(x == 2) {
			actionSendUrl(u->jid());
		}
		else if(x == 4) {
			actionStatusShow(u->jid());
		}
		else if(x == 6) {
			if(online) {
				actionAuth(u->jid());
				QMessageBox::information(d->cv, tr("Authorize"),
				tr("Sent authorization to <b>%1</b>.").arg(name));
			}
		}
		else if(x == 8) {
			if(online) {
				// if we have groups, but we are setting to 'none', then remove that particular group
				if(!u->groups().isEmpty()) {
					QString gname = ((ContactViewItem *)i->parent())->groupName();
					actionGroupRemove(u->jid(), gname);
				}
			}
		}
		else if(x == 9) {
			if(online) {
				while(1) {
					bool ok = false;
					QString newgroup = QInputDialog::getText(tr("Create New Group"), tr("Enter the new Group name:"), QLineEdit::Normal, QString::null, &ok, d->cv);
					if(!ok)
						break;
					if(newgroup.isEmpty())
						continue;

					// make sure we don't have it already
					bool found = false;
					const QStringList &groups = u->groups();
					for(QStringList::ConstIterator it = groups.begin(); it != groups.end(); ++it) {
						if(*it == newgroup) {
							found = true;
							break;
						}
					}
					if(!found) {
						QString gname = ((ContactViewItem *)i->parent())->groupName();
						actionGroupRemove(u->jid(), gname);
						actionGroupAdd(u->jid(), newgroup);
						break;
					}
				}
			}
		}
		else if(x == 10) {
			if(online) {
				actionAdd(u->jid());
				actionAuth(u->jid());

				QMessageBox::information(d->cv, tr("Add"),
				tr("Added/Authorized <b>%1</b> to the contact list.").arg(name));
			}
		}
		else if(x == 11) {
			if(online) {
				actionAuthRequest(u->jid());
				QMessageBox::information(d->cv, tr("Authorize"),
				tr("Rerequested authorization from <b>%1</b>.").arg(name));
			}
		}
		else if(x == 12) {
			actionInfo(u->jid());
		}
		else if(x == 13) {
			actionTest(u->jid());
		}
		else if(x == 15) {
			if(online) {
				int n = QMessageBox::information(d->cv, tr("Remove"),
				tr("Are you sure you want to remove authorization from <b>%1</b>?").arg(name),
				tr("&No"), tr("&Yes"));

				if(n == 1)
					actionAuthRemove(u->jid());
			}
		}
		else if(x == 16) {
			if(online)
				actionAgentSetStatus(u->jid(), STATUS_OFFLINE);
		}
		else if(x >= base_sendto && x < base_hidden) {
			int n = x - base_sendto;
			int res = n / 2;
			int type = n % 2;
			QString rname = "";
			//if(res > 0) {
				const UserResource &r = rl[res];
				rname = r.name();
			//}
			QString s = u->jid().userHost();
			if(!rname.isEmpty()) {
				s += '/';
				s += rname;
			}
			Jid j(s);

			if(type == 0)
				actionSendMessage(j);
			else if(type == 1)
				actionOpenChatSpecific(j);
		}
		else if(x >= base_hidden && x < base_group) {
			int n = 0;
			int n2 = x - base_hidden;

			QString rname;
			for(QStringList::ConstIterator it = hc.begin(); it != hc.end(); ++it) {
				if(n == n2) {
					rname = *it;
					break;
				}
				++n;
			}

			QString s = u->jid().userHost();
			if(!rname.isEmpty()) {
				s += '/';
				s += rname;
			}
			Jid j(s);

			actionOpenChatSpecific(j);
		}
		else if(x >= base_group) {
			if(online) {
				int n = 0;
				int n2 = x - base_group;

				QString newgroup;
				for(QStringList::Iterator it = gl.begin(); it != gl.end(); ++it) {
					if(n == n2) {
						newgroup = *it;
						break;
					}
					++n;
				}
				if(n == n2) {
					// remove the group of this cvi if there is one
					if(!u->groups().isEmpty()) {
						QString gname = ((ContactViewItem *)i->parent())->groupName();
						actionGroupRemove(u->jid(), gname);
					}
					// add the group
					actionGroupAdd(u->jid(), newgroup);
				}
			}
		}
	}
}

void ContactProfile::scActionDefault(ContactViewItem *i)
{
	if(i->type() == ContactViewItem::Contact)
		actionDefault(i->u()->jid());
}

void ContactProfile::scRecvEvent(ContactViewItem *i)
{
	if(i->type() == ContactViewItem::Contact)
		actionRecvEvent(i->u()->jid());
}

void ContactProfile::scSendMessage(ContactViewItem *i)
{
	if(i->type() == ContactViewItem::Contact)
		actionSendMessage(i->u()->jid());
}

void ContactProfile::scRename(ContactViewItem *i)
{
	if(!d->pa->loggedIn())
		return;

	if((i->type() == ContactViewItem::Contact && i->u()->inList()) || (i->type() == ContactViewItem::Group && i->groupType() == ContactViewItem::gUser)) {
		i->setRenameEnabled(0, true);
		i->startRename(0);
		i->setRenameEnabled(0, false);
	}
}

void ContactProfile::scHistory(ContactViewItem *i)
{
	if(i->type() == ContactViewItem::Contact)
		actionHistory(i->u()->jid());
}

void ContactProfile::scOpenChat(ContactViewItem *i)
{
	if(i->type() == ContactViewItem::Contact)
		actionOpenChat(i->u()->jid());
}

void ContactProfile::scAgentSetStatus(ContactViewItem *i, int s)
{
	if(i->type() != ContactViewItem::Contact)
		return;
	if(!i->isAgent())
		return;

	if(i->u()->isAvailable() || !d->pa->loggedIn())
		return;

	actionAgentSetStatus(i->u()->jid(), s);
}

void ContactProfile::scRemove(ContactViewItem *i)
{
	if(i->type() != ContactViewItem::Contact)
		return;

	Entry *e = findEntry(i);
	if(!e)
		return;

	bool ok = true;
	if(!d->pa->loggedIn())
		ok = false;
	if(!i->u()->inList())
		ok = true;

	if(ok) {
		QString name = e->u.jid().full();
		QString show = jidnick(e->u.jid().full(), e->u.name());
		if(name != show)
			name += QString(" (%1)").arg(e->u.name());

		int n = 0;
		if(i->parentGroupType() != ContactViewItem::gNotInList) {
			n = QMessageBox::information(d->cv, tr("Remove"),
			tr("Are you sure you want to remove <b>%1</b> from your contact list?").arg(name),
			tr("&No"), tr("&Yes"));
		}
		else
			n = 1;

		if(n == 1) {
			Jid j = e->u.jid();
			removeEntry(e);
			actionRemove(j);
		}
	}
}

void ContactProfile::doItemRenamed(ContactViewItem *i, const QString &text)
{
	if(i->type() == ContactViewItem::Contact) {
		Entry *e = findEntry(i);
		if(!e)
			return;

		// no change?
		//if(text == i->text(0))
		//	return;
		if(text.isEmpty()) {
			i->resetName();
			QMessageBox::information(d->cv, tr("Error"), tr("You can't set a blank name."));
			return;
		}

		//e->u.setName(text);
		//i->setContact(&e->u);
		actionRename(e->u.jid(), text);
	}
	else {
		// no change?
		if(text == i->groupName()) {
			i->resetGroupName();
			return;
		}
		if(text.isEmpty()) {
			i->resetGroupName();
			QMessageBox::information(d->cv, tr("Error"), tr("You can't set a blank group name."));
			return;
		}

		// make sure we don't have it already
		QStringList g = groupList();
		bool found = false;
		for(QStringList::ConstIterator it = g.begin(); it != g.end(); ++it) {
			if(*it == text) {
				found = true;
				break;
			}
		}
		if(found) {
			i->resetGroupName();
			QMessageBox::information(d->cv, tr("Error"), tr("You already have a group with that name."));
			return;
		}

		QString oldName = i->groupName();

		// set group name
		i->setGroupName(text);

		// send signal
		actionGroupRename(oldName, text);
	}
}

void ContactProfile::dragDrop(const QString &text, ContactViewItem *i)
{
	if(!d->pa->loggedIn())
		return;

	// get group
	ContactViewItem *gr;
	if(i->type() == ContactViewItem::Group)
		gr = i;
	else
		gr = (ContactViewItem *)i->parent();

	Jid j(text);
	if(!j.isValid())
		return;
	Entry *e = findEntry(j);
	if(!e)
		return;
	const UserListItem &u = e->u;
	QStringList gl = u.groups();

	// already in the general group
	if(gr->groupType() == ContactViewItem::gGeneral && gl.isEmpty())
		return;
	// already in this user group
	if(gr->groupType() == ContactViewItem::gUser && u.inGroup(gr->groupName()))
		return;

	//printf("putting [%s] into group [%s]\n", u.jid().full().latin1(), gr->groupName().latin1());

	// remove all other groups from this contact
	for(QStringList::ConstIterator it = gl.begin(); it != gl.end(); ++it) {
		actionGroupRemove(u.jid(), *it);
	}
	if(gr->groupType() == ContactViewItem::gUser) {
		// add the new group
		actionGroupAdd(u.jid(), gr->groupName());
	}
}


//----------------------------------------------------------------------------
// ContactView
//----------------------------------------------------------------------------
class ContactView::Private
{
public:
	Private() {}

	QPtrList<ContactProfile> profiles;
};

ContactView::ContactView(QWidget *parent, const char *name)
:QListView(parent, name), QToolTip(viewport())
{
	d = new Private;

	// setup the QListView
	setAllColumnsShowFocus(true);
	setShowToolTips(false);
	setHScrollBarMode(AlwaysOff);
	setMinimumSize(96,64);
	setTreeStepSize(4);
	setSorting(0,true);

	// create the column and hide the header
	addColumn("");
	header()->hide();

	setResizeMode(QListView::LastColumn);
	setDefaultRenameAction(QListView::Accept);

	// initialize the alerting stuff
	QTimer *t = new QTimer(this);
	connect(t, SIGNAL(timeout()), SLOT(animAlert()));
	t->start(120);

	// catch signals
	lcto_active = false;
	connect(this, SIGNAL(itemRenamed(QListViewItem *, int, const QString &)), SLOT(qlv_itemRenamed(QListViewItem *, int, const QString &)));
	connect(this, SIGNAL(mouseButtonPressed(int, QListViewItem *, const QPoint &, int)),SLOT(qlv_singleclick(int, QListViewItem *, const QPoint &, int)) );
	connect(this, SIGNAL(doubleClicked(QListViewItem *)),SLOT(qlv_doubleclick(QListViewItem *)) );

	v_showOffline = true;
	v_showAway = true;
	v_showAgents = true;

	// actions
	qa_send = new QAction("", tr("Send &message"), QListView::CTRL+QListView::Key_M, this);
	connect(qa_send, SIGNAL(activated()), SLOT(doSendMessage()));
	qa_ren = new QAction("", tr("Re&name"), QListView::Key_F2, this);
	connect(qa_ren, SIGNAL(activated()), SLOT(doRename()));
	qa_chat = new QAction("", tr("Open &chat window"), QListView::CTRL+QListView::Key_C, this);
	connect(qa_chat, SIGNAL(activated()), SLOT(doOpenChat()));
	qa_hist = new QAction("", tr("&History"), QListView::CTRL+QListView::Key_H, this);
	connect(qa_hist, SIGNAL(activated()), SLOT(doHistory()));
	qa_logon = new QAction("", tr("&Log on"), QListView::CTRL+QListView::Key_L, this);
	connect(qa_logon, SIGNAL(activated()), SLOT(doLogon()));
	qa_recv = new QAction("", tr("&Receive incoming event"), QListView::CTRL+QListView::Key_R, this);
	connect(qa_recv, SIGNAL(activated()), SLOT(doRecvEvent()));
	qa_rem = new QAction("", tr("Rem&ove"), QListView::Key_Delete, this);
	connect(qa_rem, SIGNAL(activated()), SLOT(doRemove()));

	optionsUpdate();
	setAcceptDrops(true);
	viewport()->setAcceptDrops(true);
}

ContactView::~ContactView()
{
	clear();
	delete d;
}

void ContactView::clear()
{
	d->profiles.setAutoDelete(true);
	d->profiles.clear();
	d->profiles.setAutoDelete(false);
}

void ContactView::link(ContactProfile *cp)
{
	d->profiles.append(cp);
}

void ContactView::unlink(ContactProfile *cp)
{
	d->profiles.removeRef(cp);
}

void ContactView::maybeTip(const QPoint &pos)
{
	ContactViewItem *i = (ContactViewItem *)itemAt(pos);
	if(!i)
		return;
	QRect r(itemRect(i));
	if(i->type() == ContactViewItem::Contact)
		tip(r, i->u()->makeTip());
	else
		tip(r, i->text(0) + " " + i->groupInfo());
}

void ContactView::keyPressEvent(QKeyEvent *e)
{
	if(e->key() == QListView::Key_Enter || e->key() == QListView::Key_Return)
		doEnter();
	else if(e->key() == QListView::Key_Space)
		doContext();
	else
		QListView::keyPressEvent(e);
}

void ContactView::setShowOffline(bool x)
{
	bool oldstate = v_showOffline;
	v_showOffline = x;

	if(v_showOffline != oldstate) {
		showOffline(v_showOffline);

		QPtrListIterator<ContactProfile> it(d->profiles);
		for(ContactProfile *cp; (cp = it.current()); ++it) {
			if(!v_showOffline)
				cp->removeAllUnneededContactItems();
			else
				cp->addAllNeededContactItems();
		}
	}
}

void ContactView::setShowAway(bool x)
{
	bool oldstate = v_showAway;
	v_showAway = x;

	if(v_showAway != oldstate) {
		showAway(v_showAway);

		QPtrListIterator<ContactProfile> it(d->profiles);
		for(ContactProfile *cp; (cp = it.current()); ++it) {
			if(!v_showAway)
				cp->removeAllUnneededContactItems();
			else
				cp->addAllNeededContactItems();
		}
	}
}

void ContactView::setShowAgents(bool x)
{
	bool oldstate = v_showAgents;
	v_showAgents = x;

	if(v_showAgents != oldstate) {
		showAgents(v_showAgents);

		QPtrListIterator<ContactProfile> it(d->profiles);
		for(ContactProfile *cp; (cp = it.current()); ++it) {
			if(!v_showAgents)
				cp->removeAllUnneededContactItems();
			else
				cp->addAllNeededContactItems();
		}
	}
}

// right-click context menu
void ContactView::qlv_contextPopup(QListViewItem *lvi, const QPoint &pos, int)
{
	ContactViewItem *i = (ContactViewItem *)lvi;
	if(!i)
		return;

	i->contactProfile()->doContextMenu(i, pos);
}

void ContactView::qlv_singleclick(int button, QListViewItem *i, const QPoint &pos, int c)
{
	lcto_active = false;
	QToolTip::hide();

	if(!i)
		return;

	ContactViewItem *item = (ContactViewItem *)i;
	setSelected(item, true);

	if(button == QListView::MidButton) {
		if(item->type() == ContactViewItem::Contact)
			item->contactProfile()->scActionDefault(item);
	}
	else {
		if(option.useleft) {
			if(button == QListView::LeftButton) {
				if(option.singleclick) {
					qlv_contextPopup(i, pos, c);
				}
				else {
					lcto_active = true;
					lcto_pos = pos;
					lcto_item = i;
					QTimer::singleShot(QApplication::doubleClickInterval()/2, this, SLOT(leftClickTimeOut()));
				}
			}
			else if(option.singleclick && button == QListView::RightButton) {
				if(item->type() == ContactViewItem::Contact)
					item->contactProfile()->scActionDefault(item);
			}
		}
		else {
			if(button == QListView::RightButton) {
				qlv_contextPopup(i, pos, c);
			}
			else if(button == QListView::LeftButton && option.singleclick) {
				if(item->type() == ContactViewItem::Contact)
					item->contactProfile()->scActionDefault(item);
			}
		}
	}
}

void ContactView::qlv_doubleclick(QListViewItem *i)
{
	lcto_active = false;

	if(!i)
		return;

	if(option.singleclick)
		return;

	ContactViewItem *item = (ContactViewItem *)i;
	item->contactProfile()->scActionDefault(item);
}

void ContactView::qlv_itemRenamed(QListViewItem *lvi, int, const QString &text)
{
	ContactViewItem *i = (ContactViewItem *)lvi;
	i->contactProfile()->doItemRenamed(i, text);
}

void ContactView::leftClickTimeOut()
{
	if(lcto_active) {
		QToolTip::hide();
		qlv_contextPopup(lcto_item, lcto_pos, 0);
		lcto_active = false;
	}
}

void ContactView::animAlert()
{
	// are any of the items animating?
	QListViewItemIterator it(this);
	ContactViewItem *item;

	for(; it.current() ; ++it) {
		item = (ContactViewItem *)it.current();
		if(item->isAlerting())
			item->animate();
		if(item->type() == ContactViewItem::Contact) {
			if(item->isAnimatingNick())
				item->animateNick();
		}
	}
}

void ContactView::optionsUpdate()
{
	// set the font
	QFont f;
	f.fromString(option.font[fRoster]);
	QListView::setFont(f);

	// set the text and background colors
	QPalette mypal = QListView::palette();
	mypal.setColor(QColorGroup::Text, option.color[cOnline]);
	mypal.setColor(QColorGroup::Base, option.color[cListBack]);
	QListView::setPalette(mypal);

	// reload the icons
	QListViewItemIterator it(this);
	ContactViewItem *item;
	for(; it.current() ; ++it) {
		item = (ContactViewItem *)it.current();
		item->optionsUpdate();
	}

	update();
}

void ContactView::resetAnim()
{
	for(QListViewItemIterator it(this); it.current() ; ++it) {
		ContactViewItem *item = (ContactViewItem *)it.current();
		if(item->isAlerting())
			item->resetAnim();
	}
}

void ContactView::doRecvEvent()
{
	ContactViewItem *i = (ContactViewItem *)selectedItem();
	if(!i)
		return;
	i->contactProfile()->scRecvEvent(i);
}

void ContactView::doRename()
{
	ContactViewItem *i = (ContactViewItem *)selectedItem();
	if(!i)
		return;
	i->contactProfile()->scRename(i);
}

void ContactView::doEnter()
{
	ContactViewItem *i = (ContactViewItem *)selectedItem();
	if(!i)
		return;
	i->contactProfile()->scActionDefault(i);
}

void ContactView::doContext()
{
	ContactViewItem *i = (ContactViewItem *)selectedItem();
	if(!i)
		return;
	ensureItemVisible(i);

	if(i->type() == ContactViewItem::Group)
		setOpen(i, !i->isOpen());
	else
		qlv_contextPopup(i, viewport()->mapToGlobal(QPoint(32, itemPos(i))), 0);
}

void ContactView::doSendMessage()
{
	ContactViewItem *i = (ContactViewItem *)selectedItem();
	if(!i)
		return;
	i->contactProfile()->scSendMessage(i);
}

void ContactView::doOpenChat()
{
	ContactViewItem *i = (ContactViewItem *)selectedItem();
	if(!i)
		return;
	i->contactProfile()->scOpenChat(i);
}

void ContactView::doHistory()
{
	ContactViewItem *i = (ContactViewItem *)selectedItem();
	if(!i)
		return;
	i->contactProfile()->scHistory(i);
}

void ContactView::doLogon()
{
	ContactViewItem *i = (ContactViewItem *)selectedItem();
	if(!i)
		return;
	i->contactProfile()->scAgentSetStatus(i, STATUS_ONLINE);
}

void ContactView::doRemove()
{
	ContactViewItem *i = (ContactViewItem *)selectedItem();
	if(!i)
		return;
	i->contactProfile()->scRemove(i);
}

QDragObject *ContactView::dragObject()
{
	ContactViewItem *i = (ContactViewItem *)selectedItem();
	if(!i)
		return 0;
	if(i->type() != ContactViewItem::Contact)
		return 0;

	QDragObject *d = new QTextDrag(i->u()->jid().full(), this);
	d->setPixmap(is.online, QPoint(8,8));
	return d;
}


//----------------------------------------------------------------------------
// ContactViewItem
//----------------------------------------------------------------------------
class ContactViewItem::Private
{
public:
	Private() {}

	int type;
	ContactProfile *cp;

	// profiles
	QString profileName;
	bool ssl;

	// groups
	int groupType;
	QString groupName;
	QString groupInfo;

	// contact
	UserListItem *u;
	int status;
	bool isAgent;
	bool alerting;
	bool animatingNick;
	Anim *anim;
	int animStep;
	int animateNickX, animateNickColor;
};

ContactViewItem::ContactViewItem(const QString &profileName, ContactProfile *cp, ContactView *parent)
:QListViewItem(parent)
{
	d = new Private;
	d->cp = cp;
	d->type = Profile;
	d->profileName = profileName;
	d->alerting = false;
	d->ssl = false;

	setProfileState(STATUS_OFFLINE);
	setText(0, profileName);
	setOpen(true);
}

ContactViewItem::ContactViewItem(const QString &groupName, int groupType, ContactProfile *cp, ContactView *parent)
:QListViewItem(parent)
{
	d = new Private;
	d->cp = cp;
	d->type = Group;
	d->groupName = groupName;
	d->groupType = groupType;
	d->alerting = false;

	drawGroupIcon();
	resetGroupName();
	setOpen(true);
	setDropEnabled(true);
}

ContactViewItem::ContactViewItem(const QString &groupName, int groupType, ContactProfile *cp, ContactViewItem *parent)
:QListViewItem(parent)
{
	if(!parent->isVisible())
		setVisible(false);

	d = new Private;
	d->cp = cp;
	d->type = Group;
	d->groupName = groupName;
	d->groupType = groupType;
	d->alerting = false;

	drawGroupIcon();
	resetGroupName();
	setOpen(true);
	setDropEnabled(true);
}

ContactViewItem::ContactViewItem(UserListItem *u, ContactProfile *cp, ContactViewItem *parent)
:QListViewItem(parent)
{
	if(!parent->isVisible())
		setVisible(false);

	d = new Private;
	d->cp = cp;
	d->type = Contact;
	d->u = u;
	d->alerting = false;
	d->animatingNick = false;

	cacheValues();

	resetStatus();
	resetName();

	setDragEnabled(true);
	setDropEnabled(true);
}

ContactViewItem::~ContactViewItem()
{
	delete d;
}

void ContactViewItem::cacheValues()
{
	if(!d->u->isAvailable())
		d->status = STATUS_OFFLINE;
	else
		d->status = makeSTATUS((*d->u->priority()).status());
	d->isAgent = d->u->isTransport();
}

ContactProfile *ContactViewItem::contactProfile() const
{
	return d->cp;
}

int ContactViewItem::type() const
{
	return d->type;
}

const QString & ContactViewItem::groupName() const
{
	return d->groupName;
}

const QString & ContactViewItem::groupInfo() const
{
	return d->groupInfo;
}

int ContactViewItem::groupType() const
{
	return d->groupType;
}

UserListItem *ContactViewItem::u() const
{
	return d->u;
}

int ContactViewItem::status() const
{
	return d->status;
}

bool ContactViewItem::isAgent() const
{
	return d->isAgent;
}

bool ContactViewItem::isAlerting() const
{
	return d->alerting;
}

bool ContactViewItem::isAnimatingNick() const
{
	return d->animatingNick;
}

int ContactViewItem::parentGroupType() const
{
	ContactViewItem *item = (ContactViewItem *)parent();
	return item->groupType();
}

void ContactViewItem::drawGroupIcon()
{
	if(d->type == Group) {
		if(childCount() == 0)
			setPixmap(0, is.arrow[2]);
		else if(isOpen())
			setPixmap(0, is.arrow[1]);
		else
			setPixmap(0, is.arrow[0]);
	}
}

void ContactViewItem::paintFocus(QPainter *, const QColorGroup &, const QRect &)
{
	// re-implimented to do nothing.  selection is enough of a focus
}

void ContactViewItem::paintBranches(QPainter *p, const QColorGroup &cg, int w, int, int h)
{
	// paint a square of nothing
	p->fillRect(0, 0, w, h, cg.base());
}

void ContactViewItem::paintCell(QPainter *p, const QColorGroup & cg, int column, int width, int alignment)
{
	if(d->type == Contact) {
		QColorGroup xcg = cg;

		if(d->status == STATUS_AWAY || d->status == STATUS_XA)
			xcg.setColor(QColorGroup::Text, option.color[cAway]);
		else if(d->status == STATUS_DND)
			xcg.setColor(QColorGroup::Text, option.color[cDND]);
		else if(d->status == STATUS_OFFLINE)
			xcg.setColor(QColorGroup::Text, option.color[cOffline]);

		if(d->animatingNick) {
			xcg.setColor(QColorGroup::Text, d->animateNickColor ? Qt::red : Qt::black);
			xcg.setColor(QColorGroup::HighlightedText, d->animateNickColor ? Qt::red : Qt::black);
		}

		QListViewItem::paintCell(p, xcg, column, width, alignment);
	}
	else if(d->type == Group || d->type == Profile) {
		QColorGroup xcg = cg;
		xcg.setColor(QColorGroup::Text, option.color[cGroupFore]);
		xcg.setColor(QColorGroup::Base, option.color[cGroupBack]);

		//QFont f = p->font();
		//f.setPixelSize(8);
		//p->setFont(f);

		QListViewItem::paintCell(p, xcg, column, width, alignment);

		p->setPen(QPen(option.color[cGroupFore]));
		p->drawRect(0, 0, width, height());

		QFontMetrics fm(p->font());
		int x = fm.width(text(column)) + pixmap(column)->width() + 8;

		if(d->type == Profile) {
			const QPixmap &pix = d->ssl ? is.ssl_yes : is.ssl_no;
			int y = (height() - pix.height()) / 2;
			p->drawPixmap(x, y, pix);
			x += 24;
		}

		if(isSelected())
			p->setPen(xcg.highlightedText());
		else
			p->setPen(xcg.text());

		int y = ((height() - fm.height()) / 2) + fm.ascent();
		p->drawText(x, y, d->groupInfo);
	}
}

void ContactViewItem::setOpen(bool o)
{
	QListViewItem::setOpen(o);
	drawGroupIcon();
}

void ContactViewItem::insertItem(QListViewItem *i)
{
	QListViewItem::insertItem(i);
	drawGroupIcon();
}

void ContactViewItem::takeItem(QListViewItem *i)
{
	QListViewItem::takeItem(i);
	drawGroupIcon();
}

int ContactViewItem::rankGroup(int groupType) const
{
	static int rankgroups[4] = {
		gGeneral,
		gUser,
		gAgents,
		gNotInList,
	};

	int n;
	for(n = 0; n < 4; ++n) {
		if(rankgroups[n] == groupType)
			break;
	}
	if(n == 4)
		return 3;

	return n;
}

int ContactViewItem::rankStatus(int status) const
{
	static int rankstatuses[5] = {
		STATUS_ONLINE,
		STATUS_AWAY,
		STATUS_XA,
		STATUS_DND,
		STATUS_OFFLINE,
	};

	int n;
	for(n = 0; n < 5; ++n) {
		if(rankstatuses[n] == status)
			break;
	}
	if(n == 5)
		return 4;

	return n;
}

int ContactViewItem::compare(QListViewItem *lvi, int, bool) const
{
	ContactViewItem *i = (ContactViewItem *)lvi;
	int ret = 0;

	if(d->type == Group || d->type == Profile) {
		// contacts always go before groups
		if(i->type() == Contact)
			ret = 1;
		else if(i->type() == Group) {
			ret = rankGroup(d->groupType) - rankGroup(i->groupType());
			if(ret == 0)
				ret = text(0).lower().localeAwareCompare(i->text(0).lower());
		}
		else if(i->type() == Profile) {
			ret = text(0).lower().localeAwareCompare(i->text(0).lower());
		}
	}
	else if(d->type == Contact) {
		// contacts always go before groups
		if(i->type() == Group)
			ret = -1;
		else {
			ret = rankStatus(d->status) - rankStatus(i->status());
			if(ret == 0)
				ret = text(0).lower().localeAwareCompare(i->text(0).lower());
		}
	}

	return ret;
}

void ContactViewItem::setProfileName(const QString &name)
{
	d->profileName = name;
	setText(0, d->profileName);
}

void ContactViewItem::setProfileState(int status)
{
	if(status == -1) {
		d->anim = &is.conn;
		d->alerting = true;
		d->animStep = 0;
		if(option.alertStyle == 0)
			setPixmap(0, d->anim->base());
	}
	else {
		d->alerting = false;
		setPixmap(0, is.status(status));
	}
}

void ContactViewItem::setProfileSSL(bool on)
{
	d->ssl = on;
	repaint();
}

void ContactViewItem::setGroupName(const QString &name)
{
	d->groupName = name;
	resetGroupName();

	updatePosition();
}

void ContactViewItem::setGroupInfo(const QString &info)
{
	d->groupInfo = info;
	repaint();
}

void ContactViewItem::resetStatus()
{
	if(!d->alerting) {
		if(!d->u->presenceError().isEmpty())
			setPixmap(0, is.perr);
		else if(d->u->ask() == "subscribe")
			setPixmap(0, is.ask);
		else if((d->u->subscription().type() == Subscription::From || d->u->subscription().type() == Subscription::None) && !d->u->isAvailable())
			setPixmap(0, is.noauth);
		else
			setPixmap(0, d->isAgent ? transport2icon(d->u->jid().full(), d->status) : is.status(d->status).pixmap());
	}
}

void ContactViewItem::resetName()
{
	resetStatus();
	setText(0, jidnick(d->u->jid().full(), d->u->name()));
}

void ContactViewItem::resetGroupName()
{
	setText(0, d->groupName);
}

void ContactViewItem::resetAnim()
{
	d->animStep = 0;
	if(d->alerting) {
		if(option.alertStyle == 0)
			setPixmap(0, d->anim->base());
	}
}

void ContactViewItem::animate()
{
	if(d->anim->isAnim() && option.alertStyle == 2) {
		setPixmap(0, d->anim->frame(d->animStep));
		++d->animStep;
		if(d->animStep > d->anim->numFrames()-1)
			d->animStep = 0;
	}
	else if(option.alertStyle != 0) {
		if(d->animStep == 0)
			setPixmap(0, d->anim->base());
		else if(d->animStep == 4)
			setPixmap(0, *pix_blank);

		++d->animStep;
		if(d->animStep == 8)
			d->animStep = 0;
	}
}

void ContactViewItem::animateNick()
{
	d->animateNickColor = (d->animateNickX % 10) >= 5 ? 1: 0;
	if((d->animateNickX % 5) == 0)
		repaint();

	++d->animateNickX;
	if(d->animateNickX == 80)
		stopAnimateNick();
}

void ContactViewItem::stopAnimateNick()
{
	if(!d->animatingNick)
		return;

	d->animatingNick = false;
	repaint();
}

void ContactViewItem::updatePosition()
{
	ContactViewItem *par = (ContactViewItem *)parent();
	if(!par)
		return;

	ContactViewItem *after = 0;
	for(QListViewItem *i = par->firstChild(); i; i = i->nextSibling()) {
		ContactViewItem *item = (ContactViewItem *)i;
		// skip self
		if(item == this)
			continue;
		int x = compare(item, 0, true);
		if(x == 0)
			continue;
		if(x < 0)
			break;
		after = item;
	}

	if(after)
		moveItem(after);
	else {
		QListViewItem *i = par->firstChild();
		moveItem(i);
		i->moveItem(this);
	}
}

void ContactViewItem::optionsUpdate()
{
	if(d->type == Group || d->type == Profile) {
		drawGroupIcon();
	}
	else if(d->type == Contact) {
		if(!d->alerting)
			resetStatus();
		else
			resetAnim();
	}
}

void ContactViewItem::setContact(UserListItem *u)
{
	//int oldStatus = d->status;
	QString oldName = text(0);
	//bool wasAgent = d->isAgent;

	QString newName = jidnick(u->jid().full(), u->name());

	d->u = u;
	cacheValues();

	bool needUpdate = false;
	//if(d->status != oldStatus || d->isAgent != wasAgent || !u->presenceError().isEmpty()) {
		resetStatus();
		needUpdate = true;
	//}
	if(newName != oldName) {
		resetName();
		needUpdate = true;
	}

	if(needUpdate)
		updatePosition();
}

void ContactViewItem::setAnimateNick()
{
	stopAnimateNick();

	d->animatingNick = true;
	d->animateNickX = 0;
	animateNick();
}

void ContactViewItem::setAlert(Anim *type)
{
	bool reset = false;

	if(!d->alerting) {
		d->alerting = true;
		reset = true;
	}
	else {
		if(d->anim != type)
			reset = true;
	}

	if(reset) {
		d->anim = type;
		d->animStep = 0;
		if(option.alertStyle == 0)
			setPixmap(0, d->anim->base());
	}
}

void ContactViewItem::clearAlert()
{
	if(d->alerting) {
		d->alerting = false;
		resetStatus();
	}
}

bool ContactViewItem::acceptDrop(const QMimeSource *m) const
{
	if(d->type == Profile)
		return false;
	else if(d->type == Group) {
		if(d->groupType != gGeneral && d->groupType != gUser)
			return false;
	}
	else if(d->type == Contact) {
		if(d->u->isSelf())
			return false;
		ContactViewItem *par = (ContactViewItem *)parent();
		if(par->groupType() != gGeneral && par->groupType() != gUser)
			return false;
	}

	if(!QTextDrag::canDecode(m))
		return false;
	QString str;
	if(!QTextDrag::decode(m, str))
		return false;

	return true;
}

void ContactViewItem::dragEntered()
{
	//printf("entered\n");
}

void ContactViewItem::dragLeft()
{
	//printf("left\n");
}

void ContactViewItem::dropped(QDropEvent *i)
{
	if(!acceptDrop(i))
		return;

	QString text;
	if(QTextDrag::decode(i, text))
		d->cp->dragDrop(text, this);
}

int ContactViewItem::rtti() const
{
	return 5103;
}

