/*
 * jidlink.cpp - establish a link between Jabber IDs
 * Copyright (C) 2001, 2002  Justin Karneges
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include"jidlink.h"

#include<qptrlist.h>
#include<qtimer.h>
#include"xmlcommon.h"
#include"dtcp.h"

#include<stdlib.h>

using namespace Jabber;


//----------------------------------------------------------------------------
// JidLinkHandler
//----------------------------------------------------------------------------
class JidLinkHandler::Private
{
public:
	Private() {}
};

JidLinkHandler::JidLinkHandler(QObject *parent)
:QObject(parent)
{
	d = new Private;
}

JidLinkHandler::~JidLinkHandler()
{
	delete d;
}

Jid JidLinkHandler::peer() const
{
	return Jid();
}

void JidLinkHandler::close()
{
}

bool JidLinkHandler::isConnected() const
{
	return false;
}

void JidLinkHandler::write(const QByteArray &)
{
}

QByteArray JidLinkHandler::read()
{
	return QByteArray();
}

bool JidLinkHandler::canRead() const
{
	return false;
}


//----------------------------------------------------------------------------
// JidLink
//----------------------------------------------------------------------------
class JidLink::Private
{
public:
	Private() {}

	Jid peer;
	Client *client;
	JidLinkManager *m;

	QString currentMethod;
	QStringList methods;

	QString key;
	int state;
	JidLinkHandler *jsh;
};

JidLink::JidLink(Client *client)
{
	d = new Private;
	d->client = client;
	d->m = client->jlManager();
	d->m->link(this);
	d->state = Idle;
	d->jsh = 0;
}

JidLink::~JidLink()
{
	d->m->unlink(this);

	delete d->jsh;
	delete d;
}

bool JidLink::connectToJid(const Jid &jid, const QString &key, const QString &method)
{
	// TODO: return false if this jid/key pair is in use

	d->peer = jid;
	d->key = key;

	d->state = Negotiating;

	if(method.isEmpty()) {
		d->methods = d->m->supportedMethods();
	}
	else {
		d->methods.clear();
		d->methods += method;
	}
	d->currentMethod = method;

	getStream();

	return true;
}

void JidLink::accept()
{
	// TODO: make accepting work

	//d->state = Establishing;
}

Jid JidLink::peer() const
{
	return d->peer;
}

QString JidLink::key() const
{
	return d->key;
}

QString JidLink::method() const
{
	return d->currentMethod;
}

int JidLink::state() const
{
	return d->state;
}

void JidLink::neg_finished()
{
	JT_Negotiate *j = (JT_Negotiate *)sender();
	if(j->success()) {
		d->currentMethod = j->option();
		QString dstr;  dstr.sprintf("JidLink: %s prefers [%s].\n", d->peer.full().latin1(), d->currentMethod.latin1());
		d->client->debug(dstr);
		doEstablish();
	}
	else {
		QString dstr;  dstr.sprintf("JidLink: %s doesn't support available methods.\n", d->peer.full().latin1());
		d->client->debug(dstr);
		error(ErrNeg);
	}
}

void JidLink::js_finished()
{
	JT_JidLink *j = (JT_JidLink *)sender();
	if(j->success()) {
		d->client->debug("JidLink: client accepted.\n");
		d->state = Establishing;

		if(d->currentMethod.isEmpty())
			getMethod();
		else
			doEstablish();
	}
	else {
		d->client->debug("JidLink: client refused.\n");
		d->state = Idle;
	}
}

void JidLink::getMethod()
{
	JT_Negotiate *j = new JT_Negotiate(d->client->rootTask());
	connect(j, SIGNAL(finished()), SLOT(neg_finished()));
	j->get(d->peer, "jabber:iq:jidlink", d->methods);
	j->go(true);
}

void JidLink::getStream()
{
	d->client->debug("JidLink: requesting stream...\n");

	JT_JidLink *j = new JT_JidLink(d->client->rootTask());
	connect(j, SIGNAL(finished()), SLOT(js_finished()));
	j->set(d->peer, d->key);
	j->go(true);
}

void JidLink::doEstablish()
{
	d->m->doEstablish(this);
}

void JidLink::wait(const Jid &jid, const QString &key)
{
	d->peer = jid;
	d->key = key;
	d->state = Waiting;
}

void JidLink::connectSuccess(JidLinkHandler *jsh)
{
	d->jsh = jsh;
	d->client->debug(QString("jsh type=[%1]\n").arg(d->jsh->className()));
	connect(d->jsh, SIGNAL(connectionClosed()), SLOT(jsh_connectionClosed()));

	d->state = Connected;

	QString dstr; dstr.sprintf("JidLink: %s {%s} Active.\n", d->peer.full().latin1(), d->key.latin1());
	d->client->debug(dstr);
}

void JidLink::connectError()
{
	QString dstr; dstr.sprintf("JidLink: %s {%s} method %s failed.\n", d->peer.full().latin1(), d->key.latin1(), d->currentMethod.latin1());

	// remove current method
	for(QStringList::Iterator it = d->methods.begin(); it != d->methods.end(); ++it) {
		if(d->currentMethod == *it) {
			d->methods.remove(it);
			break;
		}
	}

	if(d->methods.count() == 0) {
		// TODO: signal error
		d->state = Idle;
		QString dstr; dstr.sprintf("JidLink: %s {%s} out of methods.\n", d->peer.full().latin1(), d->key.latin1());
		d->client->debug(dstr);
		return;
	}

	getMethod();
}

void JidLink::incomingJSHandler(JidLinkHandler *jsh)
{
	connectSuccess(jsh);
}

void JidLink::jsh_connectionClosed()
{
	d->state = Idle;
}


//----------------------------------------------------------------------------
// JidLinkManager
//----------------------------------------------------------------------------
struct item_dtcp
{
	DTCPConnection *conn;
	JidLink *js;
};

class JidLinkManager::Private
{
public:
	Private() {}

	Client *client;
	JidLinkList list;
	QPtrList<item_dtcp> list_dtcp;
	JT_PushNegotiate *pneg;
	JT_PushJidLink *pjs;
};

JidLinkManager::JidLinkManager(Client *client)
{
	d = new Private;
	d->client = client;

	d->list_dtcp.setAutoDelete(true);

	d->pneg = new JT_PushNegotiate(d->client->rootTask());

	d->pjs = new JT_PushJidLink(d->client->rootTask());
	connect(d->pjs, SIGNAL(incoming(const Jid &, const QString &, const QString &)), SLOT(pjs_incoming(const Jid &, const QString &, const QString &)));
}

JidLinkManager::~JidLinkManager()
{
	delete d->pjs;
	delete d->pneg;
	delete d;
}

QStringList JidLinkManager::supportedMethods() const
{
	QStringList l;
	l += "dtcp";
	l += "dtcp-reverse";
	return l;
}

const JidLinkList *JidLinkManager::streamList() const
{
	return &d->list;
}

void JidLinkManager::pjs_incoming(const Jid &from, const QString &id, const QString &key)
{
	JidLink *s = findStream(from, key);
	if(s) {
		//d->pjs->respondError(from, id, 403, "Key in use");

		// this means the peer is trying an alternate method
		d->pjs->respondSuccess(from, id);
	}
	else {
		// TODO: error if method is unsupported
		//for(QStringList::ConstIterator it = 

		s = new JidLink(d->client);
		s->wait(from, key);

		d->pjs->respondSuccess(from, id);
	}
}

void JidLinkManager::dtcp_connected()
{
	DTCPConnection *c = (DTCPConnection *)sender();

	QString dstr; dstr.sprintf("JidLinkManager: dtcp connected [%s].\n", c->className());
	d->client->debug(dstr);

	QPtrListIterator<item_dtcp> it(d->list_dtcp);
	item_dtcp *i;
	for(; (i = it.current()); ++it) {
		if(i->conn == c)
			break;
	}
	if(!i)
		return;

	JidLink *s = i->js;
	d->list_dtcp.remove(i);

	s->connectSuccess(c);
}

void JidLinkManager::dtcp_error(int)
{
	DTCPConnection *c = (DTCPConnection *)sender();

	QString dstr; dstr.sprintf("JidLinkManager: dtcp error.\n");
	d->client->debug(dstr);

	QPtrListIterator<item_dtcp> it(d->list_dtcp);
	item_dtcp *i;
	for(; (i = it.current()); ++it) {
		if(i->conn == c)
			break;
	}
	if(!i)
		return;

	JidLink *s = i->js;
	d->list_dtcp.remove(i);

	s->connectError();
}

JidLink *JidLinkManager::findStream(const Jid &peer, const QString &key) const
{
	JidLinkListIt it(d->list);
	for(JidLink *i; (i = it.current()); ++it) {
		if(!i->peer().compare(peer))
			continue;
		if(i->key() == key)
			return i;
	}
	return 0;
}

void JidLinkManager::takeIncomingJSHandler(const QString &key, JidLinkHandler *jsh)
{
	// TODO: don't assume the handler is connected!!

	QString dstr; dstr.sprintf("JidLinkManager: incoming JidLinkHandler ready. (type=[%s])\n", jsh->className());
	d->client->debug(dstr);

	// TODO: report error if the stream is missing?  delete the handler?
	JidLink *s = findStream(jsh->peer(), key);
	if(!s)
		return;

	s->incomingJSHandler(jsh);
}

void JidLinkManager::link(JidLink *s)
{
	d->list.append(s);
}

void JidLinkManager::unlink(JidLink *s)
{
	d->list.remove(s);
}

void JidLinkManager::doEstablish(JidLink *s)
{
	QString m = s->method();

	if(m == "dtcp" || m == "dtcp-reverse") {
		QDomElement comment = d->client->doc()->createElement("comment");
		QDomElement js = d->client->doc()->createElement("jidlink");
		js.setAttribute("key", s->key());
		comment.appendChild(js);

		DTCPConnection *c = new DTCPConnection(d->client->dtcpManager());

		item_dtcp *i = new item_dtcp;
		i->conn = c;
		i->js = s;
		d->list_dtcp.append(i);

		connect(c, SIGNAL(connected()), SLOT(dtcp_connected()));
		connect(c, SIGNAL(error(int)), SLOT(dtcp_error(int)));
		c->connectToJid(s->peer(), (m == "dtcp-reverse"), comment);
	}
}


//----------------------------------------------------------------------------
// JT_JidLink
//----------------------------------------------------------------------------
JT_JidLink::JT_JidLink(Task *parent)
:Task(parent)
{
}

JT_JidLink::~JT_JidLink()
{
}

void JT_JidLink::set(const Jid &_to, const QString &key)
{
	to = _to;
	v_key = key;
	iq = createIQ(doc(), "set", to.full(), id());
	QDomElement query = doc()->createElement("query");
	query.setAttribute("xmlns", "jabber:iq:jidlink");
	iq.appendChild(query);
	query.appendChild(textTag(doc(), "key", v_key));
}

void JT_JidLink::onGo()
{
	send(iq);
}

bool JT_JidLink::take(const QDomElement &x)
{
	Jid from(x.attribute("from"));
	if(x.attribute("id") != id() || !to.compare(from))
		return false;

	if(x.attribute("type") == "result")
		setSuccess();
	else
		setError(x);

	return true;
}

Jid JT_JidLink::jid() const
{
	return to;
}


//----------------------------------------------------------------------------
// JT_PushJidLink
//----------------------------------------------------------------------------
JT_PushJidLink::JT_PushJidLink(Task *parent)
:Task(parent)
{
}

JT_PushJidLink::~JT_PushJidLink()
{
}

bool JT_PushJidLink::take(const QDomElement &e)
{
	// must be an iq-set tag
	if(e.tagName() != "iq" || e.attribute("type") != "set")
		return false;

	if(queryNS(e) != "jabber:iq:jidlink")
		return false;

	Jid from(e.attribute("from"));
	QString id = e.attribute("id");
	QDomElement q = queryTag(e);

	QString key;
	bool found;
	QDomElement k = findSubTag(q, "key", &found);
	if(!found)
		return true;
	key = tagContent(k);

	incoming(from, id, key);

	return true;
}

void JT_PushJidLink::respondSuccess(const Jid &to, const QString &id)
{
	QDomElement iq = createIQ(doc(), "result", to.full(), id);
	send(iq);
}

void JT_PushJidLink::respondError(const Jid &to, const QString &id, int code, const QString &str)
{
	QDomElement iq = createIQ(doc(), "error", to.full(), id);
	QDomElement err = textTag(doc(), "error", str);
	err.setAttribute("code", QString::number(code));
	iq.appendChild(err);
	send(iq);
}


//----------------------------------------------------------------------------
// JT_Negotiate
//----------------------------------------------------------------------------
JT_Negotiate::JT_Negotiate(Task *parent)
:Task(parent)
{
}

JT_Negotiate::~JT_Negotiate()
{
}

void JT_Negotiate::get(const Jid &_to, const QString &type, const QStringList &optionsList)
{
	to = _to;
	iq = createIQ(doc(), "get", to.full(), id());
	QDomElement query = doc()->createElement("query");
	query.setAttribute("xmlns", "jabber:iq:negotiate");
	iq.appendChild(query);
	QDomElement f = doc()->createElement("feature");
	f.setAttribute("type", type);
	query.appendChild(f);
	for(QStringList::ConstIterator it = optionsList.begin(); it != optionsList.end(); ++it)
		f.appendChild(textTag(doc(), "option", *it));
}

void JT_Negotiate::onGo()
{
	send(iq);
}

bool JT_Negotiate::take(const QDomElement &x)
{
	Jid from(x.attribute("from"));
	if(x.attribute("id") != id() || !to.compare(from))
		return false;

	if(x.attribute("type") == "result") {
		QDomElement q = queryTag(x);
		bool found;
		QDomElement f = findSubTag(q, "feature", &found);
		if(!found)
			return true;
		QDomElement o = findSubTag(f, "option", &found);
		if(!found)
			return true;
		v_option = tagContent(o);
		setSuccess();
	}
	else {
		setError(x);
	}

	return true;
}

Jid JT_Negotiate::jid() const
{
	return to;
}

QString JT_Negotiate::option() const
{
	return v_option;
}


//----------------------------------------------------------------------------
// JT_PushNegotiate
//----------------------------------------------------------------------------
JT_PushNegotiate::JT_PushNegotiate(Task *parent)
:Task(parent)
{
}

JT_PushNegotiate::~JT_PushNegotiate()
{
}

bool JT_PushNegotiate::take(const QDomElement &e)
{
	// must be an iq-set tag
	if(e.tagName() != "iq" || e.attribute("type") != "get")
		return false;

	if(queryNS(e) != "jabber:iq:negotiate")
		return false;

	Jid from(e.attribute("from"));
	QString id = e.attribute("id");
	QDomElement q = queryTag(e);

	QDomElement iq = createIQ(doc(), "result", from.full(), id);
	QDomElement query = doc()->createElement("query");
	query.setAttribute("xmlns", "jabber:iq:negotiate");
	iq.appendChild(query);

	for(QDomNode n = q.firstChild(); !n.isNull(); n = n.nextSibling()) {
		QDomElement i = n.toElement();
		if(i.isNull())
			continue;
		if(i.tagName() == "feature") {
			QString type = i.attribute("type");
			QStringList optionsList;
			for(QDomNode n = i.firstChild(); !n.isNull(); n = n.nextSibling()) {
				QDomElement i = n.toElement();
				if(i.isNull())
					continue;
				if(i.tagName() == "option")
					optionsList += tagContent(i);
			}

			QString prefer = getPreferred(type, optionsList);
			if(prefer.isEmpty()) {
				// bail with error
				iq = createIQ(doc(), "error", from.full(), id);
				QDomElement err = textTag(doc(), "error", "Unsupported");
				err.setAttribute("code", "500");
				iq.appendChild(err);
				break;
			}

			QDomElement f = doc()->createElement("feature");
			query.appendChild(f);
			f.setAttribute("type", type);
			f.appendChild(textTag(doc(), "option", prefer));
		}
	}

	send(iq);

	return true;
}

QString JT_PushNegotiate::getPreferred(const QString &type, const QStringList &optionsList)
{
	if(type == "jabber:iq:jidlink") {
		for(QStringList::ConstIterator it = optionsList.begin(); it != optionsList.end(); ++it) {
			if(*it == "dtcp")
				return *it;
			else if(*it == "dtcp-reverse")
				return *it;
		}
	}

	return "";
}
