/*
 * psiaccount.cpp - handles a Psi account
 * Copyright (C) 2001, 2002  Justin Karneges
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include"psiaccount.h"

#include<qinputdialog.h>
#include<qptrlist.h>
#include<qcstring.h>
#include<qtimer.h>
#include<qmessagebox.h>

#include"psicon.h"
#include"profiles.h"
#include"kmp/client.h"
#include"kmp/stream.h"
#include"kmp/message.h"
#include"kmp/tasks.h"
#include"kmp/dtcp.h"
#include"accountdlg.h"
#include"changepwdlg.h"
#include"xmlconsole.h"
#include"userlist.h"
#include"eventdlg.h"
#include"chatdlg.h"
#include"groupchatdlg.h"
#include"statusdlg.h"
#include"infodlg.h"
#include"adduserdlg.h"
#include"historydlg.h"
#include"servicesdlg.h"
#include"eventdb.h"
#include"jltest.h"


//----------------------------------------------------------------------------
// AccountLabel
//----------------------------------------------------------------------------
AccountLabel::AccountLabel(PsiAccount *_pa, QWidget *par, bool smode)
:QLabel(par)
{
	pa = _pa;
	simpleMode = smode;
	setFrameStyle( QFrame::Panel | QFrame::Sunken );

	updateName();
	connect(pa, SIGNAL(updatedAccount()), this, SLOT(updateName()));
	connect(pa, SIGNAL(destroyed()), this, SLOT(deleteMe()));
}

AccountLabel::~AccountLabel()
{
}

void AccountLabel::updateName()
{
	setText(simpleMode ? pa->name() : pa->nameWithJid());
}

void AccountLabel::deleteMe()
{
	delete this;
}


//----------------------------------------------------------------------------
// PsiAccount
//----------------------------------------------------------------------------
struct item_dialog2
{
	QWidget *widget;
	QString className;
	Jid jid;
};

class PsiAccount::Private
{
public:
	Private() {}

	PsiCon *psi;
	QString name;
	Client *client;
	ContactProfile *cp;
	UserAccount acc, accnext;
	Jid jid, nextJid;
	Status loginStatus;
	QPtrList<item_dialog2> dialogList;
	EventQueue *eventQueue;
	XmlConsole *xmlConsole;
	UserList userList;
	UserListItem self;
	int lastIdle;
	Status lastStatus;
	bool nickFromVCard;
};

PsiAccount::PsiAccount(const UserAccount &acc, PsiCon *parent)
:QObject(0)
{
	d = new Private;
	d->psi = parent;
	d->client = 0;
	d->cp = 0;

	v_isActive = false;
	isDisconnecting = false;
	notifyOnlineOk = false;
	doReconnect = false;
	usingAutoStatus = false;
	presenceSent = false;

	d->loginStatus = Status("", "");
	d->lastIdle = 0;
	d->lastStatus = Status("", "", 0, false);

	d->dialogList.setAutoDelete(true);
	d->eventQueue = new EventQueue(d->psi);
	d->userList.setAutoDelete(true);
	d->self = UserListItem(true);
	d->self.setSubscription(Subscription::Both);
	d->nickFromVCard = false;

	// create Jabber::Client
	d->client = new Client;
	d->client->setOSName(getOSName());
	d->client->setTimeZone(getTZString(), getTZOffset());
	d->client->setClientName(PROG_NAME);
	d->client->setClientVersion(PROG_VERSION);

	if(dtcp_use) {
		DTCPManager *dtcpman = d->client->dtcpManager();
		dtcpman->setHostList(dtcp_hostList);
		if(!dtcpman->setServer(dtcp_port))
			QMessageBox::information(0, tr("Error"), tr("Unable to bind to port %1 for DTCP").arg(dtcp_port));
	}

	connect(d->client, SIGNAL(connected()), SLOT(client_connected()));
	connect(d->client, SIGNAL(handshaken()), SLOT(client_handshaken()));
	connect(d->client, SIGNAL(error(const StreamError &)), SLOT(client_error(const StreamError &)));
	connect(d->client, SIGNAL(closeFinished()), SLOT(client_closeFinished()));
	connect(d->client, SIGNAL(authFinished(bool, int, const QString &)), SLOT(client_authFinished(bool, int, const QString &)));
	connect(d->client, SIGNAL(rosterRequestFinished(bool, int, const QString &)), SLOT(client_rosterRequestFinished(bool, int, const QString &)));
	connect(d->client, SIGNAL(rosterItemAdded(const RosterItem &)), SLOT(client_rosterItemUpdated(const RosterItem &)));
	connect(d->client, SIGNAL(rosterItemUpdated(const RosterItem &)), SLOT(client_rosterItemUpdated(const RosterItem &)));
	connect(d->client, SIGNAL(rosterItemRemoved(const RosterItem &)), SLOT(client_rosterItemRemoved(const RosterItem &)));
	connect(d->client, SIGNAL(resourceAvailable(const Jid &, const Resource &)), SLOT(client_resourceAvailable(const Jid &, const Resource &)));
	connect(d->client, SIGNAL(resourceUnavailable(const Jid &, const Resource &)), SLOT(client_resourceUnavailable(const Jid &, const Resource &)));
	connect(d->client, SIGNAL(presenceError(const Jid &, int, const QString &)), SLOT(client_presenceError(const Jid &, int, const QString &)));
	connect(d->client, SIGNAL(messageReceived(const Message &)), SLOT(client_messageReceived(const Message &)));
	connect(d->client, SIGNAL(subscription(const Jid &, const QString &)), SLOT(client_subscription(const Jid &, const QString &)));
	connect(d->client, SIGNAL(debugText(const QString &)), SLOT(client_debugText(const QString &)));
	connect(d->client, SIGNAL(groupChatJoined(const Jid &)), SLOT(client_groupChatJoined(const Jid &)));
	connect(d->client, SIGNAL(groupChatLeft(const Jid &)), SLOT(client_groupChatLeft(const Jid &)));
	connect(d->client, SIGNAL(groupChatPresence(const Jid &, const Status &)), SLOT(client_groupChatPresence(const Jid &, const Status &)));
	connect(d->client, SIGNAL(groupChatError(const Jid &, int, const QString &)), SLOT(client_groupChatError(const Jid &, int, const QString &)));
	if(dtcp_use)
		connect(d->client, SIGNAL(incomingDTCPConnection(DTCPConnection *)), SLOT(client_incomingDTCPConnection(DTCPConnection *)));

	// contactprofile context
	d->cp = new ContactProfile(this, acc.name, d->psi->contactView());
	connect(d->cp, SIGNAL(actionDefault(const Jid &)),SLOT(actionDefault(const Jid &)));
	connect(d->cp, SIGNAL(actionRecvEvent(const Jid &)),SLOT(actionRecvEvent(const Jid &)));
	connect(d->cp, SIGNAL(actionSendMessage(const Jid &)),SLOT(actionSendMessage(const Jid &)));
	connect(d->cp, SIGNAL(actionSendMessage(const JidList &)),SLOT(actionSendMessage(const JidList &)));
	connect(d->cp, SIGNAL(actionSendUrl(const Jid &)),SLOT(actionSendUrl(const Jid &)));
	connect(d->cp, SIGNAL(actionRemove(const Jid &)),SLOT(actionRemove(const Jid &)));
	connect(d->cp, SIGNAL(actionRename(const Jid &, const QString &)),SLOT(actionRename(const Jid &, const QString &)));
	connect(d->cp, SIGNAL(actionGroupRename(const QString &, const QString &)),SLOT(actionGroupRename(const QString &, const QString &)));
	connect(d->cp, SIGNAL(actionHistory(const Jid &)),SLOT(actionHistory(const Jid &)));
	connect(d->cp, SIGNAL(actionStatusShow(const Jid &)),SLOT(actionStatusShow(const Jid &)));
	connect(d->cp, SIGNAL(actionOpenChat(const Jid &)),SLOT(actionOpenChat(const Jid &)));
	connect(d->cp, SIGNAL(actionOpenChatSpecific(const Jid &)),SLOT(actionOpenChatSpecific(const Jid &)));
	connect(d->cp, SIGNAL(actionAgentSetStatus(const Jid &, int)),SLOT(actionAgentSetStatus(const Jid &, int)));
	connect(d->cp, SIGNAL(actionInfo(const Jid &)),SLOT(actionInfo(const Jid &)));
	connect(d->cp, SIGNAL(actionAuth(const Jid &)),SLOT(actionAuth(const Jid &)));
	connect(d->cp, SIGNAL(actionAuthRequest(const Jid &)),SLOT(actionAuthRequest(const Jid &)));
	connect(d->cp, SIGNAL(actionAuthRemove(const Jid &)),SLOT(actionAuthRemove(const Jid &)));
	connect(d->cp, SIGNAL(actionAdd(const Jid &)),SLOT(actionAdd(const Jid &)));
	connect(d->cp, SIGNAL(actionGroupAdd(const Jid &, const QString &)),SLOT(actionGroupAdd(const Jid &, const QString &)));
	connect(d->cp, SIGNAL(actionGroupRemove(const Jid &, const QString &)),SLOT(actionGroupRemove(const Jid &, const QString &)));
	connect(d->cp, SIGNAL(actionTest(const Jid &)),SLOT(actionTest(const Jid &)));

	// restore cached roster
	for(Roster::ConstIterator it = acc.roster.begin(); it != acc.roster.end(); ++it)
		client_rosterItemUpdated(*it);

	setUserAccount(acc);

	d->psi->link(this);

	d->psi->setToggles(d->acc.tog_offline, d->acc.tog_away, d->acc.tog_agents);

	// auto-login ?
	if(d->acc.opt_auto)
		setStatus(Status("", "", d->acc.priority));

	//printf("PsiAccount: [%s] loaded\n", name().latin1());

	d->xmlConsole = new XmlConsole(this);
}

PsiAccount::~PsiAccount()
{
	logout(true);
	QString str = name();

	// nuke all related dialogs
	deleteAllDialogs();

	delete d->cp;
	delete d->client;
	delete d->eventQueue;

	d->psi->unlink(this);
	delete d;

	//printf("PsiAccount: [%s] unloaded\n", str.latin1());
}

bool PsiAccount::isActive() const
{
	return v_isActive;
}

bool PsiAccount::isConnected() const
{
	return d->client->isConnected();
}

const QString & PsiAccount::name() const
{
	return d->acc.name;
}

const UserAccount & PsiAccount::userAccount() const
{
	d->psi->getToggles(&d->acc.tog_offline, &d->acc.tog_away, &d->acc.tog_agents);

	// save the roster
	d->acc.roster.clear();
	UserListIt it(d->userList);
	for(UserListItem *u; (u = it.current()); ++it) {
		if(u->inList())
			d->acc.roster += *u;
	}

	return d->acc;
}

UserList *PsiAccount::userList() const
{
	return &d->userList;
}

Client *PsiAccount::client() const
{
	return d->client;
}

ContactProfile *PsiAccount::contactProfile() const
{
	return d->cp;
}

EventQueue *PsiAccount::eventQueue() const
{
	return d->eventQueue;
}

EDB *PsiAccount::edb() const
{
	return d->psi->edb();
}

PsiCon *PsiAccount::psi() const
{
	return d->psi;
}

void PsiAccount::setUserAccount(const UserAccount &acc)
{
	d->acc = acc;
	if(!Stream::isSSLSupported()) {
		d->acc.opt_ssl = false;
		if(d->acc.port == 5223)
			d->acc.port = 5222;
	}

	if(d->acc.opt_keepAlive)
		d->client->setNoopTime(55000);  // prevent NAT timeouts every 3 minutes
	else
		d->client->setNoopTime(0);

	d->cp->setName(d->acc.name);

	Jid j = acc.user + '@' + acc.vhost;
	d->nextJid = j;
	if(!isActive()) {
		d->jid = j;
		d->self.setJid(j);
		if(!j.user().isEmpty())
			d->cp->updateSelf(d->self);
		else
			d->cp->removeSelf();
	}
	if(!d->nickFromVCard)
		setNick(acc.user);

	updatedAccount();
}

const Jid & PsiAccount::jid() const
{
	return d->jid;
}

QString PsiAccount::nameWithJid() const
{
	return (name() + " (" + jid().full() + ')');
}

// logs on with the active account settings
void PsiAccount::login()
{
	if(isActive() && !doReconnect)
		return;

	d->jid = d->nextJid;

	v_isActive = true;
	isDisconnecting = false;
	notifyOnlineOk = false;
	presenceSent = false;

	stateChanged();

	StreamProxy p;
	p.setType(d->acc.proxy_type);
	p.setHost(d->acc.proxy_host);
	p.setPort(d->acc.proxy_port);
	if(!d->acc.proxy_user.isEmpty()) {
		p.setUseAuth(true);
		p.setUser(d->acc.proxy_user);
		p.setPass(d->acc.proxy_pass);
	}
	d->client->setProxy(p);
	d->client->setSSLEnabled(d->acc.opt_ssl);
	d->client->connectToHost(d->acc.host, d->acc.port, d->acc.opt_vhost ? d->acc.vhost : QString(""));
}

// disconnect or stop reconnecting
void PsiAccount::logout(bool fast)
{
	if(!isActive())
		return;

	// cancel reconnect
	doReconnect = false;

	if(loggedIn()) {
		// send logout status
		Status s("", "Logged out", 0, false);
		d->client->setPresence(s);
	}

	isDisconnecting = true;

	if(!fast)
		simulateRosterOffline();
	v_isActive = false;
	stateChanged();

	// disconnect
	if(d->client->isActive()) {
		d->client->close(fast);
	}

	disconnected();
}

bool PsiAccount::loggedIn() const
{
	return (v_isActive && presenceSent);
}

void PsiAccount::client_connected()
{
	//printf("PsiAccount: [%s] connected\n", name().latin1());
	stateChanged();
}

void PsiAccount::client_handshaken()
{
	//printf("PsiAccount: [%s] handshaken\n", name().latin1());

	// authorize
	if(d->acc.opt_plain)
		d->client->authPlain(d->acc.user, d->acc.pass, d->acc.resource);
	else
		d->client->authDigest(d->acc.user, d->acc.pass, d->acc.resource);
}

void PsiAccount::client_error(const StreamError &err)
{
	//printf("PsiAccount: [%s] error: %s\n", name().latin1(), err.toString().latin1());

	if(!err.isWarning()) {
		isDisconnecting = true;
		simulateRosterOffline();
	}

	// Auto-Reconnect?
	if(err.type() != StreamError::SSL && err.type() != StreamError::Proxy && d->acc.opt_reconn) {
		// reconnect in 5 seconds
		doReconnect = true;
		QTimer::singleShot(5000, this, SLOT(reconnect()));
		return;
	}

	if(err.isWarning()) {
		if(!d->acc.opt_ignoreSSLWarnings) {
			int n = QMessageBox::warning(0, CAP(tr("Warning")), err.toString(), tr("Ignore"), tr("Disconnect"));
			if(n == 0) {
				d->client->continueAfterWarning();
				return;
			}
			else {
				logout();
			}
		}
		else {
			d->client->continueAfterWarning();
			return;
		}
	}
	else {
		v_isActive = false;
		isDisconnecting = true;
		stateChanged();
		disconnected();

		QString str;
		switch(err.type()) {
			case StreamError::DNS:
				str = tr("Unable to resolve hostname.");
				break;
			case StreamError::Refused:
				str = tr("Connection refused.");
				break;
			case StreamError::Timeout:
				str = tr("Connection timeout.");
				break;
			case StreamError::Socket:
				str = tr("Socket communication error.");
				break;
			case StreamError::Disconnected:
				str = tr("Disconnected.");
				break;
			case StreamError::Handshake:
				str = tr("Error during server handshake.");
				break;
			case StreamError::SSL:
				str = tr("SSL error: %1").arg(err.details());
				break;
			case StreamError::Unknown:
			default:
				str = err.details();
				break;
		}

		if(str.isEmpty())
			str = tr("None");

		QMessageBox::critical(0, CAP(tr("Error")), tr("There was an error communicating with the Jabber server.\nDetails: %1").arg(str));
	}
}

void PsiAccount::client_closeFinished()
{
	//printf("PsiAccount: [%s] connection closed\n", name().latin1());
}

void PsiAccount::client_authFinished(bool success, int, const QString &str)
{
	if(success) {
		//printf("PsiAccount: [%s] authenticated\n", name().latin1());

		// flag roster for delete
		UserListIt it(d->userList);
		for(UserListItem *u; (u = it.current()); ++it)
			u->setFlagForDelete(true);

		// ask for roster
		d->client->rosterRequest();
	}
	else {
		//printf("PsiAccount: [%s] error authenticating: [%d, %s]\n", name().latin1(), code, str.latin1());
		logout();

		QMessageBox::critical(0, CAP(tr("Error")), tr("There was an error authenticating with the Jabber server.\nDetails: %1").arg(str));
	}
}

void PsiAccount::client_rosterRequestFinished(bool success, int, const QString &)
{
	if(success) {
		//printf("PsiAccount: [%s] roster retrieved ok.  %d entries.\n", name().latin1(), d->client->roster().count());

		// delete flagged items
		UserListIt it(d->userList);
		for(UserListItem *u; (u = it.current());) {
			if(u->flagForDelete()) {
				//QMessageBox::information(0, "blah", QString("deleting: [%1]").arg(u->jid().full()));
				d->cp->removeEntry(u->jid());
				d->userList.removeRef(u);
			}
			else
				++it;
		}
	}
	else {
		//printf("PsiAccount: [%s] error retrieving roster: [%d, %s]\n", name().latin1(), code, str.latin1());
	}

	d->client->setPresence(d->loginStatus);
	presenceSent = true;
	stateChanged();
	QTimer::singleShot(15000, this, SLOT(enableNotifyOnline()));

	JT_VCard *j = new JT_VCard(d->client->rootTask());
	connect(j, SIGNAL(finished()), SLOT(slotCheckVCard()));
	j->get(d->jid);
	j->go(true);
}

void PsiAccount::client_rosterItemUpdated(const RosterItem &r)
{
	// see if the item added is already in our local list
	UserListItem *u = d->userList.find(r.jid());
	if(u) {
		u->setFlagForDelete(false);
		u->setRosterItem(r);
	}
	else {
		// we don't have it at all, so add it
		u = new UserListItem;
		u->setRosterItem(r);
		d->userList.append(u);
	}
	u->setInList(true);

	// TODO: auto info-grab (??)
	// a new entry that is "push"?  [pushinfo]
	//if(entry->push) {
	//	VCard info;
	//	if(readUserInfo(entry->jid, &info) && !info.field[vNickname].isEmpty()) {
	//		item->nick = info.field[vNickname];
	//	}
	//	else {
	//		if(localStatus != STATUS_OFFLINE)
	//			serv->getVCard(item->jid);
	//	}
	//}

	d->cp->updateEntry(*u);
}

void PsiAccount::client_rosterItemRemoved(const RosterItem &r)
{
	UserListItem *u = d->userList.find(r.jid());
	if(!u)
		return;

	simulateContactOffline(u);

	// if the item has messages queued, then move them to 'not in list'
	if(d->eventQueue->count(r.jid()) > 0) {
		u->setInList(false);
		d->cp->updateEntry(*u);
	}
	// else remove them for good!
	else {
		d->cp->removeEntry(u->jid());
		d->userList.removeRef(u);
	}
}

void PsiAccount::client_resourceAvailable(const Jid &j, const Resource &r)
{
	bool doSound = false;
	QPtrList<UserListItem> list = findRelavent(j);
	QPtrListIterator<UserListItem> it(list);
	for(UserListItem *u; (u = it.current()); ++it) {
		bool doAnim = false;
		bool local = false;
		if(u->isSelf() && r.name() == d->client->resource())
			local = true;

		// add/update the resource
		UserResourceList::Iterator rit = u->userResourceList().find(j.resource());
		bool found = (rit == u->userResourceList().end()) ? false: true;
		if(!found) {
			UserResource ur(r);
			if(local)
				ur.setVersionString(PROG_NAME + " " + PROG_VERSION + " / " + getOSName());
			u->userResourceList() += ur;

			if(notifyOnlineOk && !local) {
				doSound = true;
				doAnim = true;
			}

			if(!local) {
				// do a client-version request (too easy!)
				JT_ClientVersion *jcv = new JT_ClientVersion(d->client->rootTask());
				connect(jcv, SIGNAL(finished()), SLOT(slotClientVersionFinished()));
				jcv->get(j);
				jcv->go(true);
			}
		}
		else {
			(*rit).setResource(r);
		}

		u->setPresenceError("");
		cpUpdate(*u, r.name(), true);

		if(doAnim && option.rosterAnim)
			d->cp->animateNick(u->jid());
	}
	if(doSound)
		playSound(option.onevent[eOnline]);
}

void PsiAccount::client_resourceUnavailable(const Jid &j, const Resource &r)
{
	bool doSound = false;
	QPtrList<UserListItem> list = findRelavent(j);
	QPtrListIterator<UserListItem> it(list);
	for(UserListItem *u; (u = it.current()); ++it) {
		bool local = false;
		if(u->isSelf() && r.name() == d->client->resource())
			local = true;

		// remove resource
		UserResourceList::Iterator rit = u->userResourceList().find(j.resource());
		bool found = (rit == u->userResourceList().end()) ? false: true;
		if(found) {
			u->setLastUnavailableStatus(r.status());
			u->userResourceList().remove(rit);

			if(!u->isAvailable()) {
				u->setLastAvailable(QDateTime::currentDateTime());
				// don't sound for our own resource
				if(!isDisconnecting && !local)
					doSound = true;
			}
		}

		u->setPresenceError("");
		cpUpdate(*u, r.name(), true);
	}
	if(doSound)
		playSound(option.onevent[eOffline]);
}

void PsiAccount::client_presenceError(const Jid &j, int, const QString &str)
{
	QPtrList<UserListItem> list = findRelavent(j);
	QPtrListIterator<UserListItem> it(list);
	for(UserListItem *u; (u = it.current()); ++it) {
		simulateContactOffline(u);
		u->setPresenceError(str);
		cpUpdate(*u, j.resource(), false);
	}
}

void PsiAccount::client_messageReceived(const Message &_m)
{
	// skip empty messages
	if(_m.body().isEmpty() && _m.urlList().isEmpty())
		return;

	// skip headlines?
	if(_m.type() == "headline" && option.ignoreHeadline)
		return;

	if(_m.type() == "groupchat") {
		GCMainDlg *w = (GCMainDlg *)dialogFind("GCMainDlg", Jid(_m.from().userHost()));
		if(w) {
			// don't play sound if it is spooled, system, or the user disabled it
			if(!option.noGCSound && !_m.spooled() && !_m.from().resource().isEmpty())
				playSound(option.onevent[eChat2]);
			w->message(_m);
		}
		return;
	}

	Message m = _m;
	if(m.type() == "error")
		m.setBody(m.errorString() + "\n------\n" + m.body());

	if(option.incomingAs == 1)
		m.setType("");
	else if(option.incomingAs == 2)
		m.setType("chat");

	// urls or subject on a chat message?  convert back to regular message
	//if(m.type() == "chat" && (!m.urlList().isEmpty() || !m.subject().isEmpty()))
	//	m.setType("");

	MessageEvent *me = new MessageEvent(m);
	me->setOriginLocal(false);
	handleEvent(me);
}

void PsiAccount::client_subscription(const Jid &j, const QString &str)
{
	// if they remove our subscription, then we lost presence
	if(str == "unsubscribed") {
		UserListItem *u = d->userList.find(j);
		if(u)
			simulateContactOffline(u);
	}

	AuthEvent *ae = new AuthEvent(j, str);
	ae->setTimeStamp(QDateTime::currentDateTime());
	handleEvent(ae);
}

void PsiAccount::client_debugText(const QString &)
{
	//printf("%s", str.latin1());
	//fflush(stdout);
}

void PsiAccount::client_incomingDTCPConnection(DTCPConnection *c)
{
	JLTestDlg *w = new JLTestDlg(c->peer(), c, this);
	w->show();
}

void PsiAccount::reconnect()
{
	if(doReconnect) {
		//printf("PsiAccount: [%s] reconnecting...\n", name().latin1());
		login();
		doReconnect = false;
	}
}

Status PsiAccount::status() const
{
	return d->loginStatus;
}

void PsiAccount::setStatus(const Status &_s)
{
	// cancel auto-status and reconnect
	usingAutoStatus = false;
	doReconnect = false;

	Status s = _s;
	s.setPriority(d->acc.priority);

	d->loginStatus = s;

	if(s.isAvailable()) {
		// if client is not active then attempt to login
		if(!isActive()) {
			Jid j = d->jid;
			if(d->acc.user.isEmpty() || d->acc.host.isEmpty() || d->acc.resource.isEmpty() || !j.isValid()) {
				QMessageBox::information(0, CAP(tr("Error")), tr("Unable to login.  Ensure your account information is filled out."));
				modify();
				return;
			}
			if(!d->acc.opt_pass) {
				bool ok = false;
				QString text = QInputDialog::getText(
					tr("Need Password"),
					QString(tr("Please enter the password for <b>%1</b>")).arg(j.full()),
					QLineEdit::Password, QString::null, &ok, 0);
				if(ok && !text.isEmpty())
					d->acc.pass = text;
				else
					return;
			}

			login();
		}
		// change status
		else {
			if(presenceSent)
				setStatusDirect(s);
		}
	}
	else {
		if(isActive())
			logout();
	}
}

void PsiAccount::setStatusDirect(const Status &_s)
{
	Status s = _s;
	s.setPriority(d->acc.priority);

	d->loginStatus = s;
	d->client->setPresence(s);
	stateChanged();
}

void PsiAccount::secondsIdle(int x)
{
	if(!loggedIn())
		return;

	int lastIdle = d->lastIdle;
	Status lastStatus = d->lastStatus;

	Resource r = *(d->client->resourceList().find(d->client->resource()));
	Status ls = r.status();

	if(ls.isAvailable()) {
		// no longer idle?
		if(lastIdle > x) {
			if(ls.isAway() && usingAutoStatus) {
				lastStatus = Status("", "", d->acc.priority);
				setStatusDirect(lastStatus);
				usingAutoStatus = false;
			}
		}
		else if( !(ls.isAway() && !usingAutoStatus) ) {
			int minutes = x / 60;

			if(option.use_asOffline && option.asOffline > 0 && minutes >= option.asOffline) {
				lastStatus = Status("", "", d->acc.priority, false);
				usingAutoStatus = false;
				logout();
			}
			else if(option.use_asXa && option.asXa > 0 && minutes >= option.asXa) {
				if(ls.show() != "xa" && lastStatus.show() != "xa") {
					lastStatus = Status("xa", option.asMessage, d->acc.priority);
					setStatusDirect(lastStatus);
					usingAutoStatus = true;
				}
			}
			else if(option.use_asAway && option.asAway > 0 && minutes >= option.asAway) {
				if(ls.show() != "away" && lastStatus.show() != "away") {
					lastStatus = Status("away", option.asMessage, d->acc.priority);
					setStatusDirect(lastStatus);
					usingAutoStatus = true;
				}
			}
		}
	}

	d->lastIdle = x;
	d->lastStatus = lastStatus;
}

void PsiAccount::playSound(const QString &str)
{
	if(str.isEmpty())
		return;

	int s = STATUS_OFFLINE;
	if(loggedIn())
		s = makeSTATUS(status());

	if(s == STATUS_DND)
		return;

	// no away sounds?
	if(option.noAwaySound && (s == STATUS_AWAY || s == STATUS_XA))
		return;

	d->psi->playSound(str);
}

QWidget *PsiAccount::dialogFind(const char *className, const Jid &j)
{
	QPtrListIterator<item_dialog2> it(d->dialogList);
	for(item_dialog2 *i; (i = it.current()); ++it) {
		// does the classname and jid match?
		if(i->className == className && i->jid.compare(j)) {
			return i->widget;
		}
	}
	return 0;
}

void PsiAccount::dialogRegister(QWidget *w, const Jid &j)
{
	item_dialog2 *i = new item_dialog2;
	i->widget = w;
	i->className = w->className();
	i->jid = j;
	d->dialogList.append(i);
}

void PsiAccount::dialogUnregister(QWidget *w)
{
	QPtrListIterator<item_dialog2> it(d->dialogList);
	for(item_dialog2 *i; (i = it.current()); ++it) {
		if(i->widget == w) {
			d->dialogList.removeRef(i);
			return;
		}
	}
}

void PsiAccount::deleteAllDialogs()
{
	QPtrListIterator<item_dialog2> it(d->dialogList);
	for(item_dialog2 *i; (i = it.current());)
		delete i->widget;
	d->dialogList.clear();
}

bool PsiAccount::checkConnected(QWidget *par)
{
	if(!loggedIn()) {
		QMessageBox::information(par, CAP(tr("Error")), tr("You must be connected to the server in order to do this."));
		return false;
	}

	return true;
}

void PsiAccount::modify()
{
	AccountModifyDlg *w = (AccountModifyDlg *)dialogFind("AccountModifyDlg");
	if(w)
		bringToFront(w);
	else {
		w = new AccountModifyDlg(this, Stream::isSSLSupported(), 0);
		w->show();
	}
}

void PsiAccount::changeVCard()
{
	actionInfo(d->jid);
}

void PsiAccount::changePW()
{
	if(!checkConnected())
		return;

	ChangePasswordDlg *w = (ChangePasswordDlg *)dialogFind("ChangePasswordDlg");
	if(w)
		bringToFront(w);
	else {
		w = new ChangePasswordDlg(this);
		w->show();
	}
}

void PsiAccount::showXmlConsole()
{
	bringToFront(d->xmlConsole);
}

void PsiAccount::openAddUserDlg()
{
	if(!checkConnected())
		return;

	AddUserDlg *w = (AddUserDlg *)dialogFind("AddUserDlg");
	if(w)
		bringToFront(w);
	else {
		QStringList gl, services, names;
		UserListIt it(d->userList);
		for(UserListItem *u; (u = it.current()); ++it) {
			if(u->isTransport()) {
				services += u->jid().full();
				names += jidnick(u->jid().full(), u->name());
			}
			const QStringList &groups = u->groups();
			if(groups.isEmpty())
				continue;
			for(QStringList::ConstIterator git = groups.begin(); git != groups.end(); ++git) {
				if(qstringlistmatch(gl, *git) == -1)
					gl.append(*git);
			}
		}

		w = new AddUserDlg(services, names, gl, this);
		connect(w, SIGNAL(add(const Jid &, const QString &, const QStringList &, bool)), SLOT(dj_add(const Jid &, const QString &, const QStringList &, bool)));
		w->show();
	}
}

void PsiAccount::doManageServices()
{
	ServicesDlg *w = new ServicesDlg(d->jid.host(), this);
	connect(w, SIGNAL(signalRegister(const Jid &)), SLOT(actionRegister(const Jid &)));
	connect(w, SIGNAL(signalSearch(const Jid &)), SLOT(actionSearch(const Jid &)));
	w->show();
}

void PsiAccount::stateChanged()
{
	if(loggedIn())
		d->cp->setState(makeSTATUS(status()));
	else {
		if(isActive()) {
			d->cp->setState(-1);
			if(d->client->isConnected() && d->client->isSSLEnabled())
				d->cp->setUsingSSL(true);
		}
		else {
			d->cp->setState(STATUS_OFFLINE);
			d->cp->setUsingSSL(false);
		}
	}

	updatedActivity();
}

void PsiAccount::simulateContactOffline(UserListItem *u)
{
	UserResourceList rl = u->userResourceList();
	for(UserResourceList::ConstIterator rit = rl.begin(); rit != rl.end(); ++rit) {
		const UserResource &r = *rit;
		Jid j = u->jid();
		if(u->jid().resource().isEmpty())
			j.setResource(r.name());
		u->setPresenceError("");
		client_resourceUnavailable(j, r);
	}
}

void PsiAccount::simulateRosterOffline()
{
	UserListIt it(d->userList);
	for(UserListItem *u; (u = it.current()); ++it)
		simulateContactOffline(u);

	// self
	{
		UserListItem *u = &d->self;
		UserResourceList rl = u->userResourceList();
		for(UserResourceList::ConstIterator rit = rl.begin(); rit != rl.end(); ++rit) {
			Jid j = u->jid();
			if(u->jid().resource().isEmpty())
				j.setResource((*rit).name());
			u->setPresenceError("");
			client_resourceUnavailable(j, *rit);
		}
	}
}

void PsiAccount::enableNotifyOnline()
{
	notifyOnlineOk = true;
}

void PsiAccount::slotClientVersionFinished()
{
	JT_ClientVersion *j = (JT_ClientVersion *)sender();
	if(j->success()) {
		QPtrList<UserListItem> list = findRelavent(j->jid());
		QPtrListIterator<UserListItem> it(list);
		for(UserListItem *u; (u = it.current()); ++it) {
			UserResourceList::Iterator rit = u->userResourceList().find(j->jid().resource());
			bool found = (rit == u->userResourceList().end()) ? false: true;
			if(!found)
				continue;

			(*rit).setVersionString(j->name() + ' ' + j->version() + " / " + j->os());

			if(u->isSelf())
				d->cp->updateSelf(*u);
			else
				d->cp->updateEntry(*u);
		}
	}
}

QPtrList<UserListItem> PsiAccount::findRelavent(const Jid &j) const
{
	QPtrList<UserListItem> list;

	// self?
	if(j.compare(d->self.jid(), false))
		list.append(&d->self);
	else {
		QPtrListIterator<UserListItem> it(d->userList);
		for(UserListItem *u; (u = it.current()); ++it) {
			if(!u->jid().compare(j, false))
				continue;

			if(!u->jid().resource().isEmpty()) {
				if(u->jid().resource() != j.resource())
					continue;
			}
			list.append(u);
		}
	}

	return list;
}

UserListItem *PsiAccount::findFirstRelavent(const Jid &j) const
{
	QPtrList<UserListItem> list = findRelavent(j);
	if(list.isEmpty())
		return 0;
	else
		return list.first();
}

UserListItem *PsiAccount::find(const Jid &j) const
{
	UserListItem *u;
	if(j.compare(d->self.jid()))
		u = &d->self;
	else
		u = d->userList.find(j);

	return u;
}

void PsiAccount::cpUpdate(const UserListItem &u, const QString &rname, bool fromPresence)
{
	if(u.isSelf())
		d->cp->updateSelf(u);
	else
		d->cp->updateEntry(u);

	PsiEvent *e = d->eventQueue->peek(u.jid());
	if(e)
		d->cp->setAlert(u.jid(), &is.event2anim(e));
	else
		d->cp->clearAlert(u.jid());

	updateContact(u);
	Jid j = u.jid();
	if(!rname.isEmpty())
		j.setResource(rname);
	updateContact(j);
	updateContact(j, fromPresence);
	d->psi->updateContactGlobal(this, j);
}

QLabel *PsiAccount::accountLabel(QWidget *par, bool simpleMode)
{
	AccountLabel *al = new AccountLabel(this, par, simpleMode);
	return al;
}

EventDlg *PsiAccount::ensureEventDlg(const Jid &j)
{
	EventDlg *w = (EventDlg *)dialogFind("EventDlg", j);
	if(!w) {
		w = new EventDlg(j, this, true);
		connect(w, SIGNAL(aReadNext(const Jid &)), SLOT(processReadNext(const Jid &)));
		connect(w, SIGNAL(aReply(const Jid &, const QString &)), SLOT(dj_composeMessage(const Jid &, const QString &)));
		connect(w, SIGNAL(aAuth(const Jid &)), SLOT(dj_addAuth(const Jid &)));
		connect(w, SIGNAL(aDeny(const Jid &)), SLOT(dj_deny(const Jid &)));
		connect(d->psi, SIGNAL(emitOptionsUpdate()), w, SLOT(optionsUpdate()));
		connect(this, SIGNAL(updateContact(const Jid &)), w, SLOT(updateContact(const Jid &)));
	}

	return w;
}

ChatDlg *PsiAccount::ensureChatDlg(const Jid &j)
{
	ChatDlg *c = (ChatDlg *)dialogFind("ChatDlg", j);
	if(!c) {
		// create the chatbox
		c = new ChatDlg(j, this);
		connect(c, SIGNAL(aSend(const Message &)), SLOT(dj_sendMessage(const Message &)));
		connect(c, SIGNAL(messagesRead(const Jid &)), SLOT(chatMessagesRead(const Jid &)));
		connect(c, SIGNAL(aInfo(const Jid &)), SLOT(actionInfo(const Jid &)));
		connect(c, SIGNAL(aHistory(const Jid &)), SLOT(actionHistory(const Jid &)));
		connect(d->psi, SIGNAL(emitOptionsUpdate()), c, SLOT(optionsUpdate()));
		connect(this, SIGNAL(updateContact(const Jid &, bool)), c, SLOT(updateContact(const Jid &, bool)));
	}

	return c;
}

void PsiAccount::changeStatus(int x)
{
	if(x == STATUS_OFFLINE) {
		setStatus(Status("","",0,false));
	}
	else {
		if(x == STATUS_ONLINE && !option.askOnline) {
			setStatus(Status());
		}
		else {
			StatusSetDlg *w = new StatusSetDlg(this, makeStatus(x, ""));
			connect(w, SIGNAL(set(const Status &)), SLOT(setStatus(const Status &)));
			w->show();
		}
	}
}

void PsiAccount::actionTest(const Jid &j)
{
	UserListItem *u = find(j);
	if(!u)
		return;
	Jid j2 = j;
	if(j.resource().isEmpty()) {
		if(u->isAvailable())
			j2.setResource((*u->userResourceList().priority()).name());
	}

	JLTestDlg *w = new JLTestDlg(j2, 0, this);
	w->show();
}

void PsiAccount::actionDefault(const Jid &j)
{
	UserListItem *u = find(j);
	if(!u)
		return;

	if(d->eventQueue->count(u->jid()) > 0)
		openNextEvent(*u);
	else {
		if(option.defaultAction == 0)
			actionSendMessage(u->jid());
		else
			actionOpenChat(u->jid());
	}
}

void PsiAccount::actionRecvEvent(const Jid &j)
{
	UserListItem *u = find(j);
	if(!u)
		return;

	openNextEvent(*u);
}

void PsiAccount::actionSendMessage(const Jid &j)
{
	EventDlg *w = d->psi->createEventDlg(j.full(), this);
	w->show();
}

void PsiAccount::actionSendMessage(const JidList &j)
{
	QString str;
	bool first = true;
	for(QValueList<Jid>::ConstIterator it = j.begin(); it != j.end(); ++it) {
		if(!first)
			str += ", ";
		first = false;

		str += (*it).full();
	}

	EventDlg *w = d->psi->createEventDlg(str, this);
	w->show();
}

void PsiAccount::actionSendUrl(const Jid &j)
{
	EventDlg *w = d->psi->createEventDlg(j.full(), this);
	w->setUrlOnShow();
	w->show();
}

void PsiAccount::actionRemove(const Jid &j)
{
	dj_remove(j);
}

void PsiAccount::actionRename(const Jid &j, const QString &name)
{
	dj_rename(j, name);
}

void PsiAccount::actionGroupRename(const QString &oldname, const QString &newname)
{
	UserListIt it(d->userList);
	QPtrList<UserListItem> nu;
	for(UserListItem *u; (u = it.current()); ++it) {
		if(u->inGroup(oldname)) {
			u->removeGroup(oldname);
			u->addGroup(newname);
			cpUpdate(*u);
			if(u->inList())
				nu.append(u);
		}
	}

	if(!nu.isEmpty()) {
		JT_Roster *r = new JT_Roster(d->client->rootTask());

		QPtrListIterator<UserListItem> it(nu);
		for(UserListItem *u; (u = it.current()); ++it)
			r->set(u->jid(), u->name(), u->groups());

		r->go(true);
	}
}

void PsiAccount::actionHistory(const Jid &j)
{
	HistoryDlg *w = (HistoryDlg *)dialogFind("HistoryDlg", j);
	if(w)
		bringToFront(w);
	else {
		w = new HistoryDlg(j, this);
		connect(w, SIGNAL(openEvent(PsiEvent *)), SLOT(actionHistoryBox(PsiEvent *)));
		w->show();
	}
}

void PsiAccount::actionHistoryBox(PsiEvent *e)
{
	EventDlg *w = new EventDlg(e->from(), this, false);
	connect(w, SIGNAL(aReply(const Jid &, const QString &)), SLOT(dj_composeMessage(const Jid &, const QString &)));
	connect(w, SIGNAL(aAuth(const Jid &)), SLOT(dj_addAuth(const Jid &)));
	connect(w, SIGNAL(aDeny(const Jid &)), SLOT(dj_deny(const Jid &)));
	connect(d->psi, SIGNAL(emitOptionsUpdate()), w, SLOT(optionsUpdate()));
	connect(this, SIGNAL(updateContact(const Jid &)), w, SLOT(updateContact(const Jid &)));
	w->updateEvent(e);
	w->show();
}

void PsiAccount::actionStatusShow(const Jid &j)
{
	UserListItem *u = find(j);
	if(!u)
		return;

	StatusShowDlg *w = new StatusShowDlg(*u);
	w->show();
}

void PsiAccount::actionOpenChat(const Jid &j)
{
	UserListItem *u = find(j);
	if(!u)
		return;
	Jid j2 = j;
	if(j.resource().isEmpty()) {
		if(u->isAvailable())
			j2.setResource((*u->userResourceList().priority()).name());
	}

	openChat(j2);
}

void PsiAccount::actionOpenChatSpecific(const Jid &j)
{
	openChat(j);
}

void PsiAccount::actionAgentSetStatus(const Jid &j, int x)
{
	JT_Presence *p = new JT_Presence(d->client->rootTask());
	p->pres(j, makeStatus(x, ""));
	p->go(true);
}

void PsiAccount::actionInfo(const Jid &j)
{
	InfoDlg *w = (InfoDlg *)dialogFind("InfoDlg", j);
	if(w)
		bringToFront(w);
	else {
		/*Info *p = InfoBank::get(jid);
		VCard tmp;
		if(p)
			tmp = p->vcard;*/

		w = new InfoDlg(j.compare(d->jid) ? InfoDlg::Self : InfoDlg::Contact, j, VCard(), this, 0);
		w->show();

		// automatically retrieve info if it doesn't exist
		//if(!p)
		if(loggedIn())
			w->doRefresh();
	}
}

void PsiAccount::actionAuth(const Jid &j)
{
	dj_auth(j);
}

void PsiAccount::actionAuthRequest(const Jid &j)
{
	dj_authReq(j);
}

void PsiAccount::actionAuthRemove(const Jid &j)
{
	dj_deny(j);
}

void PsiAccount::actionAdd(const Jid &j)
{
	dj_addAuth(j);
}

void PsiAccount::actionGroupAdd(const Jid &j, const QString &g)
{
	UserListItem *u = d->userList.find(j);
	if(!u)
		return;

	if(!u->addGroup(g))
		return;
	cpUpdate(*u);

	JT_Roster *r = new JT_Roster(d->client->rootTask());
	r->set(u->jid(), u->name(), u->groups());
	r->go(true);
}

void PsiAccount::actionGroupRemove(const Jid &j, const QString &g)
{
	UserListItem *u = d->userList.find(j);
	if(!u)
		return;

	if(!u->removeGroup(g))
		return;
	cpUpdate(*u);

	JT_Roster *r = new JT_Roster(d->client->rootTask());
	r->set(u->jid(), u->name(), u->groups());
	r->go(true);
}

void PsiAccount::actionRegister(const Jid &j)
{
	if(!checkConnected())
		return;

	RegistrationDlg *w = (RegistrationDlg *)dialogFind("RegistrationDlg", j);
	if(w)
		bringToFront(w);
	else {
		w = new RegistrationDlg(j, this);
		w->show();
	}
}

void PsiAccount::actionSearch(const Jid &j)
{
	if(!checkConnected())
		return;

	SearchDlg *w = (SearchDlg *)dialogFind("SearchDlg", j);
	if(w)
		bringToFront(w);
	else {
		w = new SearchDlg(j, this);
		connect(w, SIGNAL(add(const Jid &, const QString &, const QStringList &, bool)), SLOT(dj_add(const Jid &, const QString &, const QStringList &, bool)));
		connect(w, SIGNAL(aInfo(const Jid &)), SLOT(actionInfo(const Jid &)));
		w->show();
	}
}

void PsiAccount::dj_sendMessage(const Message &m)
{
	d->client->sendMessage(m);

	// don't log groupchat messages
	if(m.type() != "groupchat") {
		MessageEvent *me = new MessageEvent(m);
		me->setOriginLocal(true);
		me->setTimeStamp(QDateTime::currentDateTime());
		logEvent(m.to(), me);
		delete me;
	}

	playSound(option.onevent[eSend]);

	// auto close an open messagebox (if non-chat)
	if(m.type() != "chat") {
		UserListItem *u = findFirstRelavent(m.to());
		if(u) {
			EventDlg *e = (EventDlg *)dialogFind("EventDlg", u->jid());
			if(e)
				e->closeAfterReply();
		}
	}
}

void PsiAccount::dj_composeMessage(const Jid &j, const QString &str)
{
	EventDlg *w = d->psi->createEventDlg(j.full(), this);
	if(!str.isEmpty())
		w->setText(qstrquote(str));
	w->show();
}

void PsiAccount::dj_addAuth(const Jid &j)
{
	QString name;
	QStringList groups;
	UserListItem *u = d->userList.find(j);
	if(u) {
		name = u->name();
		groups = u->groups();
	}

	dj_add(j, name, groups, true);
	dj_auth(j);
}

void PsiAccount::dj_add(const Jid &j, const QString &name, const QStringList &groups, bool authReq)
{
	JT_Roster *r = new JT_Roster(d->client->rootTask());
	r->set(j, name, groups);
	r->go(true);

	if(authReq)
		dj_authReq(j);
}

void PsiAccount::dj_authReq(const Jid &j)
{
	d->client->sendSubscription(j, "subscribe");
}

void PsiAccount::dj_auth(const Jid &j)
{
	d->client->sendSubscription(j, "subscribed");
}

void PsiAccount::dj_deny(const Jid &j)
{
	d->client->sendSubscription(j, "unsubscribed");
}

void PsiAccount::dj_rename(const Jid &j, const QString &name)
{
	UserListItem *u = d->userList.find(j);
	if(!u)
		return;

	QString str;
	if(name == u->jid().full())
		str = "";
	else
		str = name;

	// strange workaround to avoid a null string ??
	QString uname;
	if(u->name().isEmpty())
		uname = "";
	else
		uname = u->name();

	if(uname == str)
		return;
	u->setName(str);

	cpUpdate(*u);

	if(u->inList()) {
		JT_Roster *r = new JT_Roster(d->client->rootTask());
		r->set(u->jid(), u->name(), u->groups());
		r->go(true);
	}
}

void PsiAccount::dj_remove(const Jid &j)
{
	UserListItem *u = d->userList.find(j);
	if(!u)
		return;

	// remove all events from the queue
	d->eventQueue->clear(j);
	updateReadNext(j);

	// TODO: delete the item immediately (to simulate local change)
	if(!u->inList()) {
		simulateContactOffline(u);
		d->userList.removeRef(u);
	}
	else {
		JT_Roster *r = new JT_Roster(d->client->rootTask());
		r->remove(j);
		r->go(true);
	}
}

// handle an incoming event
void PsiAccount::handleEvent(PsiEvent *e)
{
	// find someone to accept the event
	Jid j;
	QPtrList<UserListItem> ul = findRelavent(e->from());
	if(ul.isEmpty())
		j = e->from().userHost();
	else
		j = ul.first()->jid();
	e->setJid(j);

	if(d->acc.opt_log)
		logEvent(e->from(), e);

	if(e->type() == PsiEvent::Message) {
		MessageEvent *me = (MessageEvent *)e;
		const Message &m = me->message();

		// pass chat messages directly to a chat window if possible (and deal with sound)
		if(m.type() == "chat") {
			ChatDlg *c = (ChatDlg *)dialogFind("ChatDlg", e->from());
			if(!c)
				c = (ChatDlg *)dialogFind("ChatDlg", e->jid());

			if(c && !c->isHidden()) {
				c->setJid(e->from());
				c->incomingMessage(m);
				playSound(option.onevent[eChat2]);

				if(option.alertOpenChats) {
					// to alert the chat also, we put it in the queue
					me->setSentToChatWindow(true);
					queueEvent(e);
				}
				else
					delete e;

				return;
			}

			bool firstChat = !d->eventQueue->hasChats(e->from());
			playSound(option.onevent[firstChat ? eChat1: eChat2]);
		}
		else {
			if(m.type() == "" || m.type() == "headline")
				playSound(option.onevent[eMessage]);
			else
				playSound(option.onevent[eSystem]);
		}

		if(m.type() == "error") {
			// FIXME: handle message errors
			//msg.text = QString(tr("<big>[Error Message]</big><br>%1").arg(plain2rich(msg.text)));
		}
	}
	else {
		playSound(option.onevent[eSystem]);

		AuthEvent *ae = (AuthEvent *)e;
		if(ae->authType() == "unsubscribe") {
			delete e;
			return;
		}
	}

	queueEvent(e);
}

// put an event into the event queue, and update the related alerts
void PsiAccount::queueEvent(PsiEvent *e)
{
	// do we have roster item for this?
	UserListItem *u = find(e->jid());
	if(!u) {
		// create item
		u = new UserListItem;
		u->setJid(e->jid());
		u->setInList(false);

		// treat it like a push  [pushinfo]
		//VCard info;
		//if(readUserInfo(item->jid, &info) && !info.field[vNickname].isEmpty())
		//	item->nick = info.field[vNickname];
		//else {
		//	if(localStatus != STATUS_OFFLINE)
		//		serv->getVCard(item->jid);
		//}

		d->userList.append(u);
	}

	//printf("queuing message from [%s] for [%s].\n", e->from().full().latin1(), e->jid().full().latin1());
	d->eventQueue->enqueue(e);

	// TODO: write queue to disk

	updateReadNext(e->jid());
	if(option.raise)
		d->psi->raiseMainwin();

	// udpate the roster
	cpUpdate(*u);

	bool noPopup = false;
	if(d->loginStatus.isAvailable()) {
		QString show = d->loginStatus.show();
		if(show == "dnd")
			noPopup = true;
		else if((show == "away" || show == "xa") && option.noAwayPopup)
			noPopup = true;
	}

	bool isChat = false;
	if(e->type() == PsiEvent::Message) {
		MessageEvent *me = (MessageEvent *)e;
		const Message &m = me->message();
		if(m.type() == "chat")
			isChat = true;
	}

	if(((!isChat && option.popupMsgs) || (isChat && option.popupChats)) && !noPopup) {
		UserListItem *u = find(e->jid());
		if(!u)
			return;
		openNextEvent(*u);
	}
}

// take the next event from the queue and display it
void PsiAccount::openNextEvent(const UserListItem &u)
{
	PsiEvent *e = d->eventQueue->peek(u.jid());
	if(!e)
		return;

	bool isChat = false;
	bool sentToChatWindow = false;
	if(e->type() == PsiEvent::Message) {
		MessageEvent *me = (MessageEvent *)e;
		const Message &m = me->message();
		if(m.type() == "chat") {
			isChat = true;
			sentToChatWindow = me->sentToChatWindow();
		}
	}

	if(isChat) {
		if(option.alertOpenChats && sentToChatWindow) {
			// Message already displayed, need only to pop up chat dialog, so that
			// it will be read (or marked as read)
			ChatDlg *c = (ChatDlg *)dialogFind("ChatDlg", e->from());
			if(!c)
				c = (ChatDlg *)dialogFind("ChatDlg", e->jid());
			if(!c)
				return; // should never happen

			processChats(e->from());
			bringToFront(c);
		}
		else
			openChat(e->from());
	}
	else {
		// search for an already opened eventdlg
		EventDlg *w = (EventDlg *)dialogFind("EventDlg", u.jid());

		if(!w) {
			// create the eventdlg
			w = ensureEventDlg(u.jid());

			// load the message
			processReadNext(u);

			w->show();
		}

		bringToFront(w);
	}
}

void PsiAccount::openNextEvent()
{
	PsiEvent *e = d->eventQueue->peekNext();
	if(!e)
		return;
	UserListItem *u = find(e->jid());
	if(!u)
		return;
	openNextEvent(*u);
}

void PsiAccount::updateReadNext(const Jid &j)
{
	// update eventdlg's read-next
	EventDlg *w = (EventDlg *)dialogFind("EventDlg", j);
	if(w) {
		Anim *nextAnim = 0;
		int nextAmount = d->eventQueue->count(j);
		if(nextAmount > 0)
			nextAnim = &is.event2anim(d->eventQueue->peek(j));
		w->updateReadNext(nextAnim, nextAmount);
	}

	queueChanged();
}

void PsiAccount::processReadNext(const Jid &j)
{
	UserListItem *u = find(j);
	if(u)
		processReadNext(*u);
}

void PsiAccount::processReadNext(const UserListItem &u)
{
	EventDlg *w = (EventDlg *)dialogFind("EventDlg", u.jid());
	if(!w) {
		// this should NEVER happen
		return;
	}

	// peek the event
	PsiEvent *e = d->eventQueue->peek(u.jid());
	if(!e)
		return;

	bool isChat = false;
	if(e->type() == PsiEvent::Message) {
		MessageEvent *me = (MessageEvent *)e;
		const Message &m = me->message();
		if(m.type() == "chat")
			isChat = true;
	}

	// if it's a chat message, just open the chat window.  there is no need to do
	// further processing.  the chat window will remove it from the queue, update
	// the cvlist, etc etc.
	if(isChat) {
		openChat(e->from());
		return;
	}

	// remove from queue
	e = d->eventQueue->dequeue(u.jid());

	// update the eventdlg
	w->updateEvent(e);
	delete e;

	// update the contact
	cpUpdate(u);

	updateReadNext(u.jid());
}

void PsiAccount::processChats(const Jid &j)
{
	//printf("processing chats for [%s]\n", j.full().latin1());
	ChatDlg *c = (ChatDlg *)dialogFind("ChatDlg", j);
	if(!c)
		return;

	// extract the chats
	QPtrList<PsiEvent> chatList;
	d->eventQueue->extractChats(&chatList, j);
	chatList.setAutoDelete(true);

	if(!chatList.isEmpty()) {
		// dump the chats into the chat window, and remove the related cvlist alerts
		QPtrListIterator<PsiEvent> it(chatList);
		for(PsiEvent *e; (e = it.current()); ++it) {
			MessageEvent *me = (MessageEvent *)e;
			const Message &m = me->message();

			// process the message
			if(!me->sentToChatWindow())
				c->incomingMessage(m);
		}

		QPtrList<UserListItem> ul = findRelavent(j);
		if(!ul.isEmpty()) {
			UserListItem *u = ul.first();
			cpUpdate(*u);
			updateReadNext(u->jid());
		}
	}
}

void PsiAccount::openChat(const Jid &j)
{
	ChatDlg *c = ensureChatDlg(j);
	processChats(j);
	bringToFront(c);
}

void PsiAccount::chatMessagesRead(const Jid &j)
{
	if(option.alertOpenChats)
		processChats(j);
}

void PsiAccount::logEvent(const Jid &j, PsiEvent *e)
{
	EDBHandle *h = new EDBHandle(d->psi->edb());
	connect(h, SIGNAL(finished()), SLOT(edb_finished()));
	h->append(j, e);
}

void PsiAccount::edb_finished()
{
	EDBHandle *h = (EDBHandle *)sender();
	delete h;
}

void PsiAccount::openGroupChat(const Jid &j)
{
	GCMainDlg *w = new GCMainDlg(this, j);
	connect(w, SIGNAL(aSend(const Message &)), SLOT(dj_sendMessage(const Message &)));
	connect(d->psi, SIGNAL(emitOptionsUpdate()), w, SLOT(optionsUpdate()));
	w->show();
}

bool PsiAccount::groupChatJoin(const QString &host, const QString &room, const QString &nick)
{
	return d->client->groupChatJoin(host, room, nick);
}

void PsiAccount::groupChatSetStatus(const QString &host, const QString &room, const Status &s)
{
	d->client->groupChatSetStatus(host, room, s);
}

void PsiAccount::groupChatLeave(const QString &host, const QString &room)
{
	d->client->groupChatLeave(host, room);
}

void PsiAccount::client_groupChatJoined(const Jid &j)
{
	d->client->groupChatSetStatus(j.host(), j.user(), d->loginStatus);

	GCMainDlg *m = (GCMainDlg *)dialogFind("GCMainDlg", Jid(j.userHost()));
	if(m) {
		m->joined();
		return;
	}
	GCJoinDlg *w = (GCJoinDlg *)dialogFind("GCJoinDlg", j);
	if(!w)
		return;
	w->joined();

	openGroupChat(j);
}

void PsiAccount::client_groupChatLeft(const Jid &)
{
	// TODO: anything?  nothing?  do we care?
}

void PsiAccount::client_groupChatPresence(const Jid &j, const Status &s)
{
	GCMainDlg *w = (GCMainDlg *)dialogFind("GCMainDlg", Jid(j.userHost()));
	if(!w)
		return;

	w->presence(j.resource(), s);
}

void PsiAccount::client_groupChatError(const Jid &j, int code, const QString &str)
{
	GCMainDlg *w = (GCMainDlg *)dialogFind("GCMainDlg", Jid(j.userHost()));
	if(w) {
		w->error(code, str);
	}
	else {
		GCJoinDlg *w = (GCJoinDlg *)dialogFind("GCJoinDlg", j);
		if(w) {
			w->error(code, str);
		}
	}
}

QStringList PsiAccount::hiddenChats(const Jid &j) const
{
	QStringList list;

	QPtrListIterator<item_dialog2> it(d->dialogList);
	for(item_dialog2 *i; (i = it.current()); ++it) {
		if(i->className == "ChatDlg" && i->jid.compare(j, false))
			list += i->jid.resource();
	}

	return list;
}

void PsiAccount::slotCheckVCard()
{
	JT_VCard *j = (JT_VCard *)sender();
	if(!j->success() && j->statusCode() == Task::ErrDisc) {
		setNick(d->acc.user);
		return;
	}

	if(j->vcard().isIncomplete()) {
		setNick(d->acc.user);
		changeVCard();
		return;
	}

	d->nickFromVCard = true;
	setNick(j->vcard().field[vNickname]);
}

void PsiAccount::setNick(const QString &nick)
{
	d->self.setName(nick);
	cpUpdate(d->self);
	nickChanged();
}

QString PsiAccount::nick() const
{
	return d->self.name();
}
